
# ###################################################################
# Import packages
# ###################################################################
import unittest
import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
import numpy as np
from datetime import datetime, timezone
from twisted.python import log

from martas.core import methods as mm
from martas.lib import publishing
import magpy.opt.cred as mpcred
from magpy.core import database
from magpy.core.methods import testtime


## MySQL protocol
## --------------------

class MySQLProtocol(object):
    """
    Protocol to read SQL data (usually from ttyACM0)
    MySQL protocol reads data from a MagPy database.
    All Sensors which receive continuous data updates are identified and added
    to the sensors.cfg list (marked as inactive).
    Here data can be selected and deselected. Update requires the removal of all
    data of a specific database from sensors.cfg.
    MySQL is an active protocol, requesting data at defined periods.
    """

    ## need a reference to our WS-MCU gateway factory to dispatch PubSub events
    ##
    def __init__(self, client, sensordict, confdict):
        self.client = client #self.wsMcuFactory = wsMcuFactory
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10

        self.sensorlist = []
        self.revision = self.sensordict.get('revision','')
        self.payloadformat = confdict.get("payloadformat","martas")
        try:
            self.requestrate = int(self.sensordict.get('rate','-'))
        except:
            self.requestrate = 30

        self.deltathreshold = confdict.get('timedelta')

        # debug mode
        debugtest = confdict.get('debug')
        self.debug = False
        if debugtest == 'True':
            log.msg('     DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))

        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)

        # Database specific
        self.db = self.sensor
        # get existing sensors for the relevant board
        log.msg("  -> IMPORTANT: MySQL assumes that database credentials ")
        log.msg("     are saved locally using magpy.opt.cred with the same name as database")
        try:
            self.db = database.DataBank(host=mpcred.lc(self.sensor,'host'),user=mpcred.lc(self.sensor,'user'),password=mpcred.lc(self.sensor,'passwd'),database=self.sensor)
            self.connectionMade(self.sensor)
        except:
            self.connectionLost(self.sensor,"Database could not be connected - check existance/credentials")
            return

        sensorlist = self.GetDBSensorList(self.db, searchsql='')
        self.sensor = ''
        existinglist = mm.get_sensors(confdict.get('sensorsconf'),identifier='$')

        # if there is a sensor in existinglist which is not an active sensor, then drop it
        for sensdict in existinglist:
            if sensdict.get('sensorid','') in sensorlist:
                self.sensorlist.append(sensdict)

        self.lastt = [None]*len(self.sensorlist)


    def connectionMade(self, dbname):
        log.msg('  -> Database {} connected.'.format(dbname))

    def connectionLost(self, dbname, reason=''):
        log.msg('  -> Database {} lost/not connectect. ({})'.format(dbname,reason))
        # implement counter and add three reconnection events here


    def GetDBSensorList(self, db, searchsql=''):
        """
         DESCRIPTION:
             Will connect to data base and download all data id's satisfying
             searchsql and containing data less then 5*sampling rate old.

        PARAMETER:
            existinglist: [list] [[1,2,...],['BM35_xxx_0001','SH75_xxx_0001',...]]
                          idnum is stored in sensordict['path'] (like ow)
        """

        now = datetime.now(timezone.utc).replace(tzinfo=None)

        senslist1, senslist2, senslist3 = [],[],[]
        # 1. Get search criteria (group and dataid):
        searchdataid = 'DataID LIKE "%{}"'.format(self.sensordict.get('revision',''))
        searchgroup = 'SensorGroup LIKE "%{}%"'.format(self.sensordict.get('sensorgroup',''))
        # 2. Perfom search for DataID:
        senslist1 = self.db.select('SensorID', 'DATAINFO', searchdataid)
        if self.debug:
            log.msg("  -> DEBUG - Search DATAID {}: Found {} tables".format(self.sensordict.get('revision',''),len(senslist1)))
        # 3. Perfom search for group:
        senslist2 = self.db.select('SensorID', 'SENSORS', searchgroup)
        if self.debug:
            log.msg("  -> DEBUG - Searching for GROUP {}: Found {} tables".format(self.sensordict.get('sensorgroup',''),len(senslist2)))
        # 4. Combine searchlists
        senslist = list(set(senslist1).intersection(senslist2))
        if self.debug:
            log.msg("  -> DEBUG - Fullfilling both search criteria: Found {} tables".format(len(senslist)))

        # 5. Check tables with above search criteria for recent data:
        for sens in senslist:
            datatable = sens + "_" + self.sensordict.get('revision','')
            lasttime = self.db.select('time',datatable,expert="ORDER BY time DESC LIMIT 1")
            try:
                lt = testtime(lasttime[0])
                delta = now-lt
                if self.debug:
                    log.msg("  -> DEBUG - Sensor {}: Timediff = {} sec from now, threshold={}".format(sens, delta.total_seconds(),self.deltathreshold))
                if delta.total_seconds() < float(self.deltathreshold):
                    senslist3.append(sens)
            except:
                if self.debug:
                    log.msg("  -> DEBUG - No data table?")
                pass

        # 6. Obtaining relevant sensor data for each table
        log.msg("  -> Appending sensor information to sensors.cfg")
        for sens in senslist3:
            values = {}
            values['sensorid'] = sens
            values['protocol'] = 'MySQL'
            values['port'] = '-'
            cond = 'SensorID = "{}"'.format(sens)
            vals = self.db.select('SensorName,SensorID,SensorSerialNum,SensorRevision,SensorGroup,SensorDescription,SensorTime','SENSORS',condition=cond)[0]
            vals = ['-' if el==None else el for el in vals]
            values['serialnumber'] = vals[2]
            values['name'] = vals[0]
            values['revision'] = vals[3]
            values['mode'] = 'active'
            pier = self.db.select('DataPier','DATAINFO',condition=cond)[0]
            values['pierid'] = pier
            values['ptime'] = vals[6]
            values['sensorgroup'] = vals[4]
            values['sensordesc'] = vals[5].replace(',',';')

            success = mm.add_sensors(self.confdict.get('sensorsconf'), values, block='SQL')

        return senslist3


    def sendRequest(self):
        """
        source:mysql:
        Method to obtain data from table
        """
        t1 = datetime.now(timezone.utc).replace(tzinfo=None)
        outdate = datetime.strftime(t1, "%Y-%m-%d")
        bufferfilename = outdate

        if self.debug:
            log.msg("  -> DEBUG - Sending periodic request ...")

        def getList(sql):
            cursor = self.db.db.cursor()
            keys=[]
            try:
                cursor.execute(sql)
            except:
                log.msg("  -> ERROR - get SQL data")
                cursor.close()
                return keys
            head = cursor.fetchall()
            keys = list(np.transpose(np.asarray(head))[0])
            cursor.close()
            return keys

        # get self.sensorlist
        # get last timestamps
        # read all data for each sensor since last timestamp
        # send that and store last timestamp
        for index,sensdict in enumerate(self.sensorlist):
            sensorid = sensdict.get('sensorid')
            if self.debug:
                log.msg("  -> DEBUG - dealing with sensor {}".format(sensorid))
            # 1. Getting header
            # -----------------
            # load keys, elements and units
            #header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid, key, ele, unit, multplier, packcode, struct.calcsize('<'+packcode))
            dataid = sensorid+'_'+self.revision
            keyssql = 'SHOW COLUMNS FROM %s' % (dataid)
            keystab = getList(keyssql)
            if len(keystab) > 0:
                if 'time' in keystab:
                    keystab.remove('time')
                if 'flag' in keystab:
                    keystab.remove('flag')
                if 'typ' in keystab:
                    keystab.remove('typ')
                if 'comment' in keystab:
                    keystab.remove('comment')
                keys = ','.join(keystab)
                if self.debug:
                    log.msg("  -> DEBUG - requesting header {}".format(sensorid))
                sql1 = 'SELECT SensorElements FROM SENSORS WHERE SensorID LIKE "{}"'.format(sensorid)
                sql2 = 'SELECT SensorKeys FROM SENSORS WHERE SensorID LIKE "{}"'.format(sensorid)
                sql3 = 'SELECT ColumnUnits FROM DATAINFO WHERE SensorID LIKE "{}"'.format(sensorid)
                sql4 = 'SELECT ColumnContents FROM DATAINFO WHERE SensorID LIKE "{}"'.format(sensorid)
                try:
                    elem = getList(sql1)[0].split(',')
                except:
                    elem =[]
                try:
                    keyssens = getList(sql2)[0].split(',')
                except:
                    keyssens =[]
                try:
                    unit = getList(sql3)[0].split(',')
                except:
                    unit =[]
                try:
                    cont = getList(sql4)[0].split(',')
                except:
                    cont =[]
                units, elems = [], []
                for key in keystab:
                    try:
                        pos1 = keyssens.index(key)
                        ele = elem[pos1]
                    except:
                        ele = key
                    elems.append(ele)
                    try:
                        pos2 = cont.index(ele)
                        units.append(unit[pos2])
                    except:
                        units.append('None')
                if self.debug:
                    log.msg("  -> DEBUG - creating head line {}".format(sensorid))
                multplier = '['+','.join(map(str, [10000]*len(keystab)))+']'
                packcode = '6HL'+''.join(['q']*len(keystab))
                header = ("# MagPyBin {} {} {} {} {} {} {}".format(sensorid, '['+','.join(keystab)+']', '['+','.join(elems)+']', '['+','.join(units)+']', multplier, packcode, struct.calcsize('<'+packcode)))

                if self.debug:
                    print ("DATA header", header)

                # 2. Getting dict
                sql = 'SELECT DataSamplingRate FROM DATAINFO WHERE SensorID LIKE "{}"'.format(sensorid)
                sr = float(getList(sql)[0])
                coverage = int(self.requestrate/sr)+120

                # 3. Getting data
                # get data and create typical message topic
                # based on sampling rate and collection rate -> define coverage
                data = self.db.get_lines(dataid, coverage)
                self.db.db.commit()

                #li = sorted(self.db.select('time,'+keys, dataid, expert='ORDER BY time DESC LIMIT {}'.format(int(coverage))))
                if not self.lastt[index]:
                    self.lastt[index]=data.ndarray[0][0]

                data = data.trim(starttime=self.lastt[index])
                if self.debug:
                    print ("DATA content", len(data), self.lastt[index])

                # transpose
                dataarray = data.ndarray
                orglist = [col for col in dataarray if len(col) == len(dataarray[0])]
                newli = np.transpose(orglist)

                for dataline in newli:
                    timestamp = dataline[0]
                    data_bin = None
                    datearray = ''
                    try:
                        datearray = mm.datetime_to_array(timestamp)
                        for i, para in enumerate(keystab):
                            try:
                                val = int(float(dataline[i + 1]) * 10000)
                            except:
                                val = 999990000
                            datearray.append(val)
                        data_bin = struct.pack('<' + packcode, *datearray)  # little endian
                    except:
                        log.msg('Error while packing binary data')

                    if not self.confdict.get('bufferdirectory', '') == '' and data_bin:
                        mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, bufferfilename, data_bin,
                                        header)
                    if self.debug:
                        log.msg("  -> DEBUG - sending ... {}".format(','.join(list(map(str, datearray))), header))
                    self.sendData(sensorid, ','.join(list(map(str, datearray))), header, len(newli) - 1,
                                  fullhead=data.header)

                self.lastt[index] = newli[-1][0]
                if self.debug:
                    print ("NEW STARTTIME", self.lastt[index])


        t2 = datetime.now(timezone.utc).replace(tzinfo=None)
        if self.debug:
            log.msg("  -> DEBUG - Needed {}".format(t2-t1))


    def sendData(self, sensorid, data, head, stack=None, fullhead=None):
        if not fullhead:
            fullhead = {}
        topic = self.confdict.get('station') + '/' + sensorid
        senddata = False
        if not stack:
            stack = int(self.sensordict.get('stack'))
        coll = stack

        if coll > 1:
            self.metacnt = 1 # send meta data with every block
            if self.datacnt < coll:
                self.datalst.append(data)
                self.datacnt += 1
            else:
                senddata = True
                data = ';'.join(self.datalst)
                self.datalst = []
                self.datacnt = 0
        else:
            senddata = True

        if senddata:
                if self.payloadformat == "intermagnet":
                    pubdict = publishing.intermagnet(None, topic=topic, data=data, head=head,
                                            imo=self.confdict.get('station', ''), meta=fullhead)
                else:
                    pubdict, count = publishing.martas(None, topic=topic, data=data, head=head, count=self.count,
                                            changecount=self.metacnt,
                                            imo=self.confdict.get('station', ''), meta=self.sensordict)
                    self.count = count
                for topic in pubdict:
                    if self.debug:
                        print ("Publishing", topic, pubdict.get(topic))
                    self.client.publish(topic, pubdict.get(topic), qos=self.qos)


class TestMySQLProtocol(unittest.TestCase):
    """
    Test environment for all methods
    """

    def test_init(self):
        sensordict = {'sensorid': 'cobsdb', 'port': '-', 'baudrate': '-', 'bytesize': '-', 'stopbits': '-',
                      'parity': '-', 'mode': 'active', 'init': 'None', 'rate': '120', 'stack': '1', 'protocol': 'MySQL',
                      'name': 'MySQL', 'serialnumber': '-', 'revision': '0001', 'path': '-', 'pierid': '-', 'ptime': '-',
                      'sensorgroup': 'magnetism', 'sensordesc': '-\n'}
        confdict = {'station': 'wic', 'sensorsconf': '/home/leon/.martas/conf/sensors.cfg',
                    'initdir': '/home/leon/.martas/init/', 'bufferdirectory': '/home/leon/MARTAS/mqtt',
                    'serialport': '/dev/tty', 'timeout': '1800.0', 'broker': 'localhost', 'mqttport': 1883,
                    'mqttdelay': 60, 'mqttqos': 1, 'mqttversion': 2, 'payloadformat': 'martas', 'owport': 4304,
                    'owhost': 'localhost', 'timedelta': 100, 'logging': '/home/leon/.martas/log/martas.log',
                    'debug': 'True'}
        my = MySQLProtocol(None,sensordict,confdict)
        print (my.sensorlist)



if __name__ == "__main__":
    unittest.main(verbosity=2)


#!/usr/bin/env python
# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
from datetime import datetime, timezone
from twisted.python import log
from martas.core import methods as mm
from martas.lib import publishing

from random import randint

"""
TestProtocol:
a typical sensors input would look like:
TEST_1234_0001,USB0,57600,8,1,N,active,,-,1,Test,Test,1234,0001,-,MyPier,NTP,Test environment
"""

class TestProtocol(object):
    """
    Protocol to simulate a record. Test is an active protocol.
    It will create a random number between 40 and 60 and send it at regular intervals.
    To be used for testing procedures.
    """
    def __init__(self, client, sensordict, confdict):
        """
        'client' could be used to switch between different publishing protocols
                 (e.g. MQTT or WS-MCU gateway factory) to publish events
        'sensordict' contains a dictionary with all sensor relevant data (sensors.cfg)
        'confdict' contains a dictionary with general configuration parameters (martas.cfg)
        """
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        #log.msg("  -> Sensor: {}".format(self.sensor))
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        self.qos = int(confdict.get('mqttqos',0))
        self.payloadformat = confdict.get("payloadformat","martas")
        log.msg("  -> setting MQTT publishing format:", self.payloadformat)
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)
        # debug mode
        debugtest = confdict.get('debug')
        self.debug = False
        if debugtest == 'True':
            log.msg('     DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))


    def processData(self, data):
        """Processing test data """

        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        outdate = datetime.strftime(currenttime, "%Y-%m-%d")
        filename = outdate
        timestamp = datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")
        packcode = '6hLl'
        sensorid = self.sensor
        header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid, '[x]', '[RN]', '[random]', '[1000]', packcode, struct.calcsize('<'+packcode))
        datearray = mm.time_to_array(timestamp)
        data_bin = None

        try:
            datearray.append(int(data*1000))
            data_bin = struct.pack('<'+packcode,*datearray)  #use little endian byte order
        except:
            log.msg('Error while packing binary data')
            pass

        if not self.confdict.get('bufferdirectory','') == '' and data_bin:
            mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)
        return ','.join(list(map(str,datearray))), header

    def sendRequest(self):
        topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
        value = 50. + (randint(-9, 9)) + 1./float(randint(1, 9))
        try:
            data, head = self.processData(value)

            senddata = False
            coll = int(self.sensordict.get('stack'))
            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(data)
                    self.datacnt += 1
                else:
                    senddata = True
                    data = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            pubdict = {}
            if senddata:
                if self.payloadformat == "intermagnet":
                    pubdict = publishing.intermagnet(None, topic=topic, data=data, head=head,
                                            imo=self.confdict.get('station', ''), meta=self.sensordict)
                else:
                    pubdict, count = publishing.martas(None, topic=topic, data=data, head=head, count=self.count,
                                            changecount=self.metacnt,
                                            imo=self.confdict.get('station', ''), meta=self.sensordict)
                    self.count = count
                for topic in pubdict:
                    if self.debug:
                        print ("Publishing", topic, pubdict.get(topic))
                    self.client.publish(topic, pubdict.get(topic), qos=self.qos)

        except:
            log.err('{}: Unable to parse data {}'.format(self.sensordict.get('protocol'), value))



from __future__ import print_function
from __future__ import absolute_import

# ###################################################################
# Import packages
# ###################################################################

from magpy.stream import DataStream, KEYLIST, NUMKEYLIST, subtractStreams
import struct
from datetime import datetime 
import json

def datetime2array(t):
        return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

## LORA-ZAMG - protocol
##
class lorazamg(object):
    """
    ZAMG/adeunis {"DateTime":"2018-11-28T16:17:47.517412Z","Modell":"Adeunis RF","Name":"Adeunis7 - Museumsstr","issue":{"appeui":"00-18-b2-43-45-4e-53-33","deveui":"00-18-b2-20-00-00-02-6b","code":66,"status_nr":224,"channelA mA":11.09218,"channelB mA":11.09218,"status":0},"data":{"corr":0,"tl":14.326125000000005,"rf":41.194874999999996,"msg_payload":[66,224,18,16,236,226,34,16,41,46]}}

    ZAMG/bees {"DatumSec":"2018-11-28T17:02:59.352375Z","Name":"Unknown bees","issue":{"appeui":"48-83-c7-df-30-0d-00-00","deveui":"48-83-c7-df-30-0d-12-74","breite":48.2483,"laenge":16.357428,"hoehe":198,"st":0},"data":{"tl":21.75,"rf":36.9,"saved_payload":[1,127,8,106,14,92,0]}}

    ZAMG/ikt/isd/test/jf/adeunis07_beelike {"DateTime":"2018-11-28T16:17:47.517412Z","deveui":"00-18-b2-20-00-00-02-6b","breite":48.2483,"laenge":16.357428,"hoehe":148,"status":0,"tl":14.326125000000005,"rf":41.194874999999996}



    content suggestion: appeui, deveui, sensorname, locationname, sensormodell, -> rest beelike

    topic suggestions:
    headline/station/sensor

    ideally, headline is a unique formatidentifier
    e.g.
    loraz/schwarzenbergplatz/adeunis  od 
    warum: so kann man relativ systematisch stationen abfragen
    mobile sensoren ohne festen standort:
    loraz/mobile/adeunis


    """

    def __init__(self):
        """
        """
        print ("  -> Initializing lora zamg routines ...")
        #self.payload = payload
        #self.topic = topic
        self.topicidentifier = {'startswith':'ZAMG','endswith':'s'}
        self.datakeytranslator = {'tl':['t1','degC'], 'rf':['var1','per'], 'corr':['var5','none']}
        self.identifier = {}
        self.headdict = {}
        self.headstream = {}

    def GetPayload(self, payload, topic):
        loradict = json.loads(payload)
        # convert loradict to headdict (header) and data_bin
        newpayload, sensorid, headline, header =  self.loradict2datastruct(loradict)
        return newpayload, sensorid, headline, header, self.identifier

    def loradict2datastruct(self, loradict):
            datakeytranslator = {'tl':['t1','degC'], 'rf':['var1','per'], 'corr':['var5','none']}
            datadict = loradict.get('data')
            issuedict = (loradict.get('issue'))
            header = {}
            header['SensorName'] = loradict.get('Name').split(' ')[0]
            try:
                if loradict.get('Name','').find('bee') > 0:
                    header['StationID'] = 'ZAMG bees'
                    header['StationName'] = 'ZAMG bees'
                header['StationID'] = loradict.get('Name').strip().split(' - ')[1]
                header['StationName'] = loradict.get('Name').strip().split(' - ')[1]
            except:
                header['StationID'] = "Not defined"
            #header['SensorName'] = loradict.get('Name').split(' - ').strip()[0]
            header['SensorSerialNum'] = issuedict.get('deveui','').replace('-','')
            header['SensorDataLoggerSerNum'] = issuedict.get('appeui','').replace('-','')
            header['SensorGroup'] = loradict.get('Modell','')
            sensorid = header['SensorName'].split(' ')[0] + '_' + header['SensorSerialNum'] + '_0001'
            header['SensorID'] = sensorid
            header['StationLongitude'] = issuedict.get('laenge','')
            header['StationLatitude'] = issuedict.get('breite','')
            if not issuedict.get('laenge','') == '':
                header['StationLocationReference'] = 'WGS84, EPSG: 4326'
            header['StationElevation'] = issuedict.get('hoehe','')
            if not issuedict.get('hoehe','') == '':
                header['StationElevationRef'] = issuedict.get('hoehe','')

            # needs to return headstream[sensorid] = header (global)
            # identifier dictionary  (global)
            # sensorid
            # headdict[sensorid] = "MagPy line"  (global)
            # and data payload

            keylist, elemlist, unitlist, multilist = [],[],[],[]
            if not loradict.get('DateTime','') == '':
                time = datetime.strptime(loradict.get('DateTime'),"%Y-%m-%dT%H:%M:%S.%fZ")
            elif not loradict.get('DatumSec','') == '':
                time = datetime.strptime(loradict.get('DatumSec'),"%Y-%m-%dT%H:%M:%S.%fZ")
            datalst = datetime2array(time)
            packstr = '6hL'
            for elem in datadict:
                if elem in datakeytranslator:
                    key = datakeytranslator[elem][0]
                    unit = datakeytranslator[elem][1]
                    keylist.append(key)
                    elemlist.append(elem)
                    unitlist.append(unit)
                    multilist.append(1000)
                    packstr += "l"
                    datalst.append(int(datadict[elem]*1000))   
                #print (elem, datadict[elem])

            datalst = [str(elem) for elem in datalst]
            dataline =','.join(datalst)
            #print ("DATA", dataline)
            self.identifier[sensorid+':packingcode'] = packstr
            self.identifier[sensorid+':keylist'] = keylist
            self.identifier[sensorid+':elemlist'] = elemlist
            self.identifier[sensorid+':unitlist'] = unitlist
            self.identifier[sensorid+':multilist'] = multilist

            def identifier2line(dic, sensorid):
                p1 = dic.get(sensorid+':packingcode')
                p2 = dic.get(sensorid+':keylist')
                p3 = dic.get(sensorid+':elemlist')
                p4 = dic.get(sensorid+':unitlist')
                p5 = dic.get(sensorid+':multilist')
                p5 = [str(elem) for elem in p5]
                size = struct.calcsize(p1)
                line = "# MagPyBin {} [{}] [{}] [{}] [{}] {} {}".format(sensorid,','.join(p2),','.join(p3),','.join(p4),','.join(p5),p1,size)
                return line

            headline = identifier2line(self.identifier, sensorid)
            #self.headstream[sensorid] = create_head_dict(self.headdict[sensorid],sensorid)
            #self.headstream[sensorid] = merge_two_dicts(self.headstream[sensorid], header)
            #print ("HEAD1", headdict[sensorid])
            #print ("HEAD2", headstream[sensorid])
            return dataline, sensorid, headline, header





#!/usr/bin/env python
"""
Secondary acquisition routine of MARTAS:
MQTT protocol by Roman Leonhardt and Rachel Bailey to be used in the Conrad Observatory.

How should it work:
PURPOSE:
acquisition_mqtt.py reads e.g. serial data and publishes that using the mqtt protocol.
An "collector" e.g. MARCOS can subscribe to the "publisher" and access the data stream.

REQUIREMENTS:
1.) install a MQTT broker (e.g. ubuntu (< 16.04):
sudo apt-add-repository ppa:mosquitto-dev/mosquitto-ppa
sudo apt-get update
sudo apt-get install mosquitto mosquitto-clients)
2.) Secure comm and authencation: https://www.digitalocean.com/community/tutorials/how-to-install-and-secure-the-mosquitto-mqtt-messaging-broker-on-ubuntu-16-04

METHODS:
acquisition_mqtt.py contains the following methods:

GetSensors: read a local definition file (sensors.txt) which contains information
            on SensorID, Port, Bausrate (better Serial communication details), active/passive,
            init requirements, optional SensorDesc

GetDefaults: read initialization file with local paths, publishing server, ports, etc.

SendInit: send eventually necessary initialization data as defined in sensors.txt

GetActive: Continuously obtain serial data from instrument and convert it to an binary
           information line to be published (using libraries)

GetPassive: Send scheduled request to serial port to obtain serial data from instrument
            and convert it to an binary information line to be published (using libraries)

1. how to convert incoming serial datalines to magpy.stream contents
2. an eventual initialization protocol too be send to the serial port before
3.
call method: defined here

Usage:
sudo python acquisition.py

"""

# ###################################################################
# Import packages
# ###################################################################
## Import MagPy
## -----------------------------------------------------------
#local = True
#if local:
#    import sys
#    sys.path.insert(1,'/home/leon/Software/magpy-git/')

import threading
import sys, getopt, os
from datetime import datetime, timezone
import time
import socket
import subprocess
import importlib

## Import MagPy packages
## -----------------------------------------------------------
from magpy.opt import cred as mpcred

## Import specific MARTAS packages
## -----------------------------------------------------------
sys.path.insert(1,'/home/leon/Software/MARTAS/')
from martas.core import methods as mm
from martas.version import __version__

## Import MQTT
## -----------------------------------------------------------
import paho.mqtt.client as mqtt
import paho.mqtt
import ssl

## Import twisted for serial port communication and web server
## -----------------------------------------------------------
if sys.platform == 'win32':
    ## on windows, we need to use the following reactor for serial support
    ## http://twistedmatrix.com/trac/ticket/3802
    ##
    from twisted.internet import win32eventreactor
    win32eventreactor.install()
# IMPORT TWISTED

from twisted.internet import reactor
from twisted.python import log
from twisted.internet.serialport import SerialPort


# ###################################################################
# Default specifications and initialization parameters
# ###################################################################

now = datetime.now(timezone.utc).replace(tzinfo=None)
hostname = socket.gethostname()
msgcount = 0


SUPPORTED_PROTOCOLS = {'Env' : {"name" : "Envrionement 05 sensor", "type" : "passive", "status" : "productive"},
                       'Ow' : {"name" : "One wire sensors", "status" : "productive"},
                       'Lemi' : {"name" : "LEMI sensors L25 and L36", "status" : "productive"},
                       'Arduino' : {"name" : "Arduino microcontroller", "status" : ""},
                       'GSM90' : {"name" : "GEM Systems Overhauser observatory 90", "type" : "passive (init)", "status" : "productive"},
                       'GSM19' : {"name" : "GEM Systems Overhauser mobile 19", "status" : "productive"},
                       'Cs' : {"name" : "G823 Geometrics Cesium ", "status" : "productive"},
                       'POS1' : {"name" : "Quantum POS1 Overhauser", "type" : "passive (init)", "status" : "productive"},
                       'MySQL' : {"name" : "MariaDB/MySQL database", "status" : "productive"},
                       'Lm' : {"name" : "", "status" : ""},
                       'Lnm' : {"name" : "Thiess Laser Disdrometer", "status" : ""},
                       'BM35' : {"name" : "Pressure", "type" : "passive (init)", "status" : "productive"},
                       'Test' : {"name" : "Test environment"},
                       'GP20S3' : {"name" : "GEM Systems Potassium Gradiometer", "status" : "productive"},
                       'Active' : {"name" : ""},
                       'ActiveArduino' : {"name" : "One wire sensors"},
                       'DSP' : {"name" : "Wind speed"},
                       'Disdro' : {"name" : ""},
                       'ad7714' : {"name" : "24bit ADC AD7714"},
                       'fluke289' : {"name" : "Fluke 289 Multimeter"},
                       'cr1000jc' : {"name" : ""},
                       'GIC' : {"name" : "GIC Webservice Austria TU Graz"},
                       'obsdaq' : {"name" : "Mingeo Obsdaq 24bit"},
                       'IMfile' : {"name" : "MagPy readable data files"},
                       'FourPL' : {"name" : "Lippmann 4PL geoelectric sensor", "status" : "productive"},
                       'BME280I2C' : {"name" : "BME280 T/P/rh probe on I2C", "status" : "productive"},
                       'GNSSHAT' : {"name" : "Waveshare GSM/GPRS/GNSS Hat", "status" : "develop"},
                       'MMC5603I2C' : {"name" : "MMC5603 magnetic probe on I2C", "status" : "develop"},
                       }

#SUPPORTED_PROTOCOLS = ['Env','Ow','Lemi','Arduino','GSM90','GSM19','Cs','POS1','MySQL','Lm','Lnm','BM35','Test','GP20S3','Active','ActiveArduino','DSP','Disdro','ad7714','fluke289','cr1000jc','GIC','obsdaq','imfile','FourPL','BME280I2C'] # should be provided by MagPy

"""
Protocol types:
ok		Env   		: py2,py3	: passive		: environment
ok		Ow		: py2 		: active (group)	: environment
ok		Arduino		: py2,py3	: passive (group)	: environment
ok              ActiveArduino   : py2,py3	: active (group)        : all
ok		BM35		: py2		: passive 		: environment
ok      	Lemi		: py2,py3	: passive		: mag
ok      	GSM90		: py2,py3		: passive (init)	: mag
ok      	POS1		: py2,py3	: passive (init)	: mag
written (time test missing)GSM19: py2		: passive 		: mag
ok	 	Cs		: py2,py3	: passive 		: mag
ok	   	ObsDaq 	: py2,p3		: passive		: mag
ok		MySQL		: py2		: active (group)	: general db call
-	        Active		:		: active		: general active call
current work	CR1000		: py2		: active		: all
ok 		Test 	  	: py2		: active                : random number
ok		Lnm		: py2		: active 		: environment
ok		Disdro		: py2		: active 		: environment
ok              AD7714          : py2 		: autonomous		: general ADC
current work              IMFile          : py3 		: active		: read files
current work              FourPL          : py3 		: active		: 4point light geoelectric
current work              BME280I2C       : py3 		: active		: I2C pins on raspberry with BME280 T-humidity-pressure
current work              GNSSHAT       : py3 		: active		: I2C pins on raspberry with BME280 T-humidity-pressure
current work              MMC5603I2C       : py3 		: active		: I2C pins on raspberry with BME280 T-humidity-pressure
"""

try:
    # if sslpsk2 is available import the moduls and create necessary classes
    from sslpsk2.sslpsk2 import _ssl_set_psk_server_callback, _ssl_set_psk_client_callback

    def _ssl_setup_psk_callbacks(sslobj):
        psk = sslobj.context.psk
        hint = sslobj.context.hint
        identity = sslobj.context.identity
        if psk:
            if sslobj.server_side:
                cb = psk if callable(psk) else lambda _identity: psk
                _ssl_set_psk_server_callback(sslobj, cb, hint)
            else:
                cb = psk if callable(psk) else lambda _hint: psk if isinstance(psk, tuple) else (psk, identity)
                _ssl_set_psk_client_callback(sslobj, cb)


    class SSLPSKContext(ssl.SSLContext):
        @property
        def psk(self):
            return getattr(self, "_psk", None)

        @psk.setter
        def psk(self, psk):
            self._psk = psk

        @property
        def hint(self):
            return getattr(self, "_hint", None)

        @hint.setter
        def hint(self, hint):
            self._hint = hint

        @property
        def identity(self):
            return getattr(self, "_identity", None)

        @identity.setter
        def identity(self, identity):
            self._identity = identity


    class SSLPSKObject(ssl.SSLObject):
        def do_handshake(self, *args, **kwargs):
            if not hasattr(self, '_did_psk_setup'):
                _ssl_setup_psk_callbacks(self)
                self._did_psk_setup = True
            super().do_handshake(*args, **kwargs)


    class SSLPSKSocket(ssl.SSLSocket):
        def do_handshake(self, *args, **kwargs):
            if not hasattr(self, '_did_psk_setup'):
                _ssl_setup_psk_callbacks(self)
                self._did_psk_setup = True
            super().do_handshake(*args, **kwargs)


    SSLPSKContext.sslobject_class = SSLPSKObject
    SSLPSKContext.sslsocket_class = SSLPSKSocket

except:
    pass


def SendInit(confdict,sensordict):
    """
    DESCRIPTION:
    send eventually necessary initialization data as defined in sensors.conf
    """
    pass

def do_every (interval, worker_func, iterations = 0):
    if iterations != 1:
        threading.Timer(interval,do_every, [interval, worker_func, 0 if iterations == 0 else iterations-1]).start ()
    worker_func()

def active_thread(confdict,sensordict, mqttclient, activeconnections):
    """
    1. identify protocol from sensorid
    2. Apply protocol (read serial and return data)
    3. add data to Publish
    -> do all that in while True
    """
    prot = {}
    protocol = None
    sensorid = sensordict.get('sensorid')
    log.msg("Starting active thread for {}".format(sensorid))
    protocolname = sensordict.get('protocol')
    log.msg("  -> Importing protocol {}".format(protocolname))

    protlst = [activeconnections[key] for key in activeconnections]
    amount = protlst.count(protocolname) + 1 # LoadTEST_1234 existing connections (new amount is len(exist)+1)
    if protocolname in SUPPORTED_PROTOCOLS:
        module = "martas.lib.{}protocol".format(protocolname.lower())
        cls = "{}Protocol".format(protocolname)
        pname =  "{}{}".format(protocolname, amount)
        if confdict.get('debug') == 'True':
            log.msg("DEBUG -> Importing: {} from {}".format(cls,module))
        prot[pname] = getattr(importlib.import_module(module), cls)
        if confdict.get('debug') == 'True':
            print("... importing done")
            print("Initializing the protocol ...")
        protocol = prot.get(pname)(mqttclient, sensordict, confdict)
        log.msg("... protocol successfully initialized")
    else:
        log.msg("  -> did not find protocol in SUPPORTED_PROTOCOLS")

    log.msg("  -> Starting active thread ...")

    try:
        rate = float(sensordict.get('rate'))
        log.msg("  -> using provided sampling rate of {} sec".format(rate))
    except:
        log.msg("  -> did not find appropriate sampling rate - using 30 sec")
        rate = 30.

    do_every(rate, protocol.sendRequest)

    activeconnection = {sensorid: protocolname}
    log.msg("  -> active connection established ... sampling every {} sec".format(rate))

    return activeconnection

def passive_thread(confdict,sensordict, mqttclient, establishedconnections):
    """
    1. identify protocol from sensorid
    2. Apply protocol (read serial and return data)
    3. add data to Publish
    -> do all that in while True
    """
    prot = {}
    pname = None
    protocol = None
    sensorid = sensordict.get('sensorid')
    log.msg("Starting passive thread for {}".format(sensorid))
    protocolname = sensordict.get('protocol')
    log.msg("  -> Found protocol {}".format(protocolname))
    protlst = [establishedconnections[key] for key in establishedconnections]
    amount = protlst.count(protocolname) + 1 # Load existing connections (new amount is len(exist)+1)
    #amount = 1                           # Load existing connections (new amount is len(exist)+1)
    if protocolname in SUPPORTED_PROTOCOLS:
        module = "martas.lib.{}protocol".format(protocolname.lower())
        cls = "{}Protocol".format(protocolname)
        pname =  "{}{}".format(protocolname, amount)
        if confdict.get('debug') == 'True':
            log.msg("DEBUG -> Importing: {} from {}".format(cls, module))
        prot[pname] = getattr(importlib.import_module(module), cls)
        if confdict.get('debug') == 'True':
            print("... importing done")
            print("Initializing the protocol ...")
        protocol = prot.get(pname)(mqttclient, sensordict, confdict)
        log.msg("... protocol successfully initialized")
    else:
        log.msg("  -> did not find protocol in SUPPORTED_PROTOCOLS")

    port = confdict['serialport']+sensordict.get('port')
    log.msg("  -> Connecting to port {} ...".format(port))
    if confdict.get('debug') == 'True':
        log.msg("DEBUG -> perameters - protocol {}, port {}, baudrate {}".format(pname, port, int(sensordict.get('baudrate'))))
    serialPort = SerialPort(protocol, port, reactor, baudrate=int(sensordict.get('baudrate')))

    passiveconnection = {sensorid: protocolname}
    log.msg("  -> passive connection established")

    return passiveconnection

def auto_thread(confdict,sensordict, mqttclient, establishedconnections):
    """
    1. identify protocol from sensorid
    2. Apply protocol (an autonoumous thread will be started, who publishes data)
    """
    prot = {}
    sensorid = sensordict.get('sensorid')
    log.msg("Starting auto thread for {}".format(sensorid))
    protocolname = sensordict.get('protocol')
    log.msg("  -> Found protocol {}".format(protocolname))
    protlst = [establishedconnections[key] for key in establishedconnections]
    amount = protlst.count(protocolname) + 1 # Load existing connections (new amount is len(exist)+1)
    #amount = 1                           # Load existing connections (new amount is len(exist)+1)
    if protocolname in SUPPORTED_PROTOCOLS:
        module = "martas.lib.{}protocol".format(protocolname.lower())
        cls = "{}Protocol".format(protocolname)
        pname =  "{}{}".format(protocolname, amount)
        if confdict.get('debug') == 'True':
            log.msg("DEBUG -> Importing: {} from {}".format(cls, module))
        prot[pname] = getattr(importlib.import_module(module), cls)
        if confdict.get('debug') == 'True':
            print("... importing done")
            print("Initializing the protocol ...")
        protocol = prot.get(pname)(mqttclient, sensordict, confdict)
        log.msg("... protocol successfully initialized")
    else:
        log.msg("  -> did not find protocol in SUPPORTED_PROTOCOLS")

    autoconnection = {sensorid: protocolname}
    log.msg("  ->  autonomous connection established")

    return autoconnection

# -------------------------------------------------------------------
# MQTT connect:
# -------------------------------------------------------------------

def on_connect(client, userdata, flags, rc, properties):
    log.msg("Connected with result code " + str(rc))
    global msgcount
    if rc == 0 and msgcount < 4:
        log.msg("Moving on...")
    elif rc == 5 and msgcount < 4:
        log.msg("Authentication required")
    msgcount += 1
    # add a counter here with max logs

def on_message(client, userdata, message):
   # Decode the payload to get rid of the 'b' prefix and single quotes:
   log.msg('It is ' + str(message.payload.decode("utf-8")))

def on_disconnect(client, userdata, message):
   log.msg("Disconnected from the broker.")

#####################################################################
# MAIN PROGRAM
#####################################################################

def main(argv):
    ##
    ## Run like: python acquisition.py -m '/home/cobs/MARTAS/defaults.conf'

    global now
    global hostname
    global msgcount
    global SUPPORTED_PROTOCOLS

    passive_count = 0
    active_count = 0
    martasfile = 'martas.cfg'
    cred = ''
    creduser = ''
    credhost = ''
    pskidentity = b""
    pskpwd = ""
    pwd = 'None'
    debug = False
    test = False
    conf={}

    ##  Get eventually provided options
    ##  ----------------------------
    usagestring = 'acquisition.py -m <martas>'
    try:
        opts, args = getopt.getopt(argv,"hm:TD",["martas=","test=","debug=",])
    except getopt.GetoptError:
        print('Check your options:')
        print(usagestring)
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-h':
            print('------------------------------------------------------')
            print('Usage:')
            print(usagestring)
            print('------------------------------------------------------')
            print('Options:')
            print('-h                             help')
            print('-m                             path to martas configuration')
            print('------------------------------------------------------')
            print('Examples:')
            print('1. Basic (using defauilt martas.cfg')
            print('   python acquisition.py')
            print('2. Using other configuration')
            print('   python acquisition.py -m "/home/myuser/mymartas.cfg"')
            sys.exit()
        elif opt in ("-m", "--martas"):
            martasfile = arg
        elif opt in ("-T", "--test"):
            test = True
        elif opt in ("-D", "--debug"):
            debug = True

    ##  Load defaults dict
    ##  ----------------------------
    conf = mm.get_conf(martasfile)
    conf["station"] = conf.get("station").lower()
    if conf.get('debug') in ["True","TRUE","true"]:
        debug = True

    if debug:
        print (" Running acquisition tool in debug mode")

    if test:
        # When running test then conf will be predefined
        conf["station"] = "TST".lower()
        conf["broker"] = "localhost"
        conf["bufferdirectory"] = "/tmp"
        conf["logging"] = "sys.stdout"
        conf["debug"] = "True"

    # Add a check routine here whether conf information was obtained
    if debug:
        print ("Configuration:", conf)

    cred = conf.get('mqttcred',"")
    credpath = conf.get('credentialpath', None)
    broker = conf.get('broker',"")
    mqttport = int(conf.get('mqttport',1883))
    mqttdelay = int(conf.get('mqttdelay',60))
    mqttcert = conf.get('mqttcert',"")
    mqttpsk = conf.get('mqttpsk',"")
    if mqttpsk:
        pskidentity = mpcred.lc(mqttpsk, 'user', path=credpath)
        pskpwd = mpcred.lc(mqttpsk, 'passwd', path=credpath)
    if debug:
        print ("MQTTPSK", mqttpsk, credpath)
        print ("PSKID", pskidentity)

    ##  Get Sensor data
    ##  ----------------------------
    sensorlist = mm.get_sensors(conf.get('sensorsconf',""))
    if test:
        sensorlist = [{'sensorid': 'TEST_1234_0001', 'port': 'USB0', 'baudrate': '57600', 'bytesize': '8', 'stopbits': '1', 'parity': 'N', 'mode': 'active', 'init': '', 'rate': '2', 'stack': '1', 'protocol': 'Test', 'name': 'TEST', 'serialnumber': '1234', 'revision': '0001', 'path': '-', 'pierid': 'MyPier', 'ptime': 'NTP', 'sensorgroup': 'Test environment'}]

    if debug:
        print ("Sensor data:", sensorlist)

    ## Check for credentials
    ## ----------------------------
    if not cred == '':
        try:
            print ("Accessing credential information for {}".format(cred))
            creduser = mpcred.lc(cred,'user',path=credpath)
            pwd = mpcred.lc(cred,'passwd',path=credpath)
        except:
            print ("error when accessing credentials")
            cred = ""
            pass

    ## create MQTT client
    ##  ----------------------------
    pahovers = paho.mqtt.__version__
    pahomajor = int(pahovers[0])
    client = None
    if debug:
        print (" MQTT version ", pahovers)
    if pahomajor <= 1:
        client = mqtt.Client(clean_session=True)
    else:
        mqttversion = int(conf.get("mqttversion", 2))
        if mqttversion == 2:
            print (" Importing MQTT API version 2")
            client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, clean_session=True)
        else:
            print (" Importing MQTT API version 1")
            client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION1, clean_session=True)

    if cred:
        # Should have two possibilities:
        print ("MQTT: password authentication")
        client.username_pw_set(username=creduser,password=pwd)

    if int(mqttport) >= 8883:
        if mqttcert:
            if debug:
                print("MQTT: TLS encryption based on certificate")
                print(mqttcert)
            client.tls_set(ca_certs=mqttcert)
        elif mqttpsk and pskidentity:
            if debug:
                print("MQTT: TLS encryption based on PSK - requires the sslpsk2 python module")
            # making use of discussions in https://github.com/eclipse-paho/paho.mqtt.python/issues/451
            #context = SSLPSKContext(ssl.PROTOCOL_TLS_CLIENT) # This does bot work for beaglebone (python3.11 clients)
            if debug:
                print ("MARTAS TLS: Deprecation warning - but necessary for old clients")
            context = SSLPSKContext(ssl.PROTOCOL_TLSv1_2)
            context.set_ciphers('PSK')
            context.psk = bytes.fromhex(pskpwd)
            context.identity = pskidentity.encode()
            client.tls_set_context(context)  # Here we apply the new `SSLPSKContext`
        else:
            if debug:
                print("MQTT: basic TLS")
            client.tls_set(ca_certs=None, certfile=None, keyfile=None, cert_reqs=ssl.CERT_REQUIRED, tls_version=ssl.PROTOCOL_TLS,
                    ciphers=None)

    ##  Start Twisted logging system
    ##  ----------------------------
    if conf.get('logging').strip() == 'sys.stdout':
        log.startLogging(sys.stdout)
    else:
        try:
            print (" -- Logging to {}".format(conf.get('logging')))
            log.startLogging(open(conf.get('logging'),'a'))
            log.msg("----------------")
            log.msg("  -> Logging to {}".format(conf.get('logging')))
        except:
            log.startLogging(sys.stdout)
            print ("Could not open {}. Switching log to stdout.".format(conf['logging']))

    log.msg("----------------")
    log.msg("Starting MARTAS acquisition version {}".format(__version__))
    log.msg("----------------")

    ## connect to MQTT client
    ##  ----------------------------
    client.on_connect = on_connect
    try:
        client.connect(broker, mqttport, mqttdelay)
        client.loop_start()
    except:
        log.msg("Critical error - no network connection available during startup or mosquitto server not running - check whether data is recorded")

    establishedconnections = {}
    ## Connect to serial port (sensor dependency) -> returns publish
    # Start subprocesses for each publishing protocol
    for sensor in sensorlist:
        log.msg("----------------")
        log.msg("Sensor and Mode:", sensor.get('sensorid'), sensor.get('mode'))
        log.msg("----------------")
        init = sensor.get('init')
        if not init in ['','None',None,0,'-']:
            log.msg("  - Initialization using {}".format(init))
            initdir = conf.get('initdir')
            initapp = os.path.join(initdir,init)
            # Check if provided initscript is existing
            try:
                log.msg("  - running initialization {}".format(initapp))
                # initcall = "{} {}".format(sys.executable, initapp)  # only for python scripts
                # log.msg(subprocess.check_output(['/bin/sh',initapp])) # only for shell scripts
                # first line of init script has to be #!/bin/sh or #!/bin/sh , #/usr/bin/python etc...
                log.msg(subprocess.check_output(initapp))
            except subprocess.CalledProcessError as e:
                log.msg("  - init command '{}' returned with error (code {}): {}".format(e.cmd, e.returncode, e.output))
            except:
                pass
        if sensor.get('mode') in ['p','passive','Passive','P']:
            if debug:
                # In case of debug mode do not catch exceptions with try ... except
                connected = passive_thread(conf, sensor, client, establishedconnections)
                log.msg(" - Passive thread initiated in debug mode for {}. Ready to receive data ...".format(sensor.get('sensorid')))
                establishedconnections.update(connected)
                passive_count += 1
            else:
                try:
                    connected = passive_thread(conf,sensor,client,establishedconnections)
                    log.msg(" - Passive thread initiated for {}. Ready to receive data ...".format(sensor.get('sensorid')))
                    establishedconnections.update(connected)
                    passive_count +=1
                except:
                    log.msg(" - !!! Passive thread failed for {} !!!".format(sensor.get('sensorid')))
                    pass
        elif sensor.get('mode') in ['a','active','Active','A']:
            if debug:
                # In case of debug mode do not catch exceptions with try ... except
                connected_act = active_thread(conf, sensor, client, establishedconnections)
                log.msg(" - Active thread initiated in debug mode for {}. Periodically requesting data ...".format(
                    sensor.get('sensorid')))
            else:
                try:
                    log.msg(" - Active thread initiated for {}. Periodically requesting data ...".format(sensor.get('sensorid')))
                    connected_act = active_thread(conf,sensor,client,establishedconnections)
                except:
                    log.msg(" - !!! Active thread failed for {} !!!".format(sensor.get('sensorid')))
                    pass
        elif sensor.get('mode') in ['autonomous']:
            try:
                log.msg(" - Auto thread initiated for {}. Ready to receive data ...".format(sensor.get('sensorid')))
                connected_act = auto_thread(conf,sensor,client,establishedconnections)
            except Exception as e:
                log.msg(" - !!! Auto thread failed for {} !!!".format(sensor.get('sensorid')))
                log.msg(e)
                pass
        else:
            log.msg("acquisition: Mode not recognized")

        sensorid = sensor.get('sensorid')

    # Start all passive clients
    if passive_count > 0:
        log.msg("acquisition: Starting reactor for passive sensors. Sending data now ...")
        reactor.run()

    # TODO other solution - when the main thread stops, the deamons are killed!
    got_here = True
    print("main thread running ...")
    if got_here:
        while True:
            time.sleep(100)


if __name__ == "__main__":
   main(sys.argv[1:])



"""
fluke289protocol written 2023 by richard.mandl@geosphere.at
for MARTAS MQTT protocol by Roman Leonhardt and Rachel Bailey to be used in the Conrad Observatory.
Makes sense only when using a Fluke 289 mulimeter or compatible
Example for a line in sensors.txt:
    FLUKE_14430039_0001,USB0,115200,8,1,N,active,-,1,1,fluke289,FLUKE,14430039,0001,-,-,GPS,electric,multimeter
"""
from __future__ import print_function
from __future__ import absolute_import

# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
from datetime import datetime, timedelta
from twisted.python import log
#from core import acquisitionsupport as acs
from martas.core import methods as mm
import serial
import time, sys

# ###################################################################

# some helpful function
def datetime2array(t):
    return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

# constants for Fluke 289 readings
STATE = ['INVALID','NORMAL','BLANK','DISCHARGE','OL','OL_MINUS','OPEN_TC']
ATTRIBUTE = ['NONE','OPEN_CIRCUIT','SHORT_CIRCUIT','GLITCH_CIRCUIT','GOOD_DIODE','LO_OHMS','NEGATIVE_EDGE','POSITIVE_EDGE','HIGH_CURRENT']

PREFIX = ['n','u','m','','k','M']

class fluke289Protocol:        
    """
    Protocol to read from a Fluke 289 Multimeter
    """

    def __init__(self, client, sensordict, confdict):
        self.client = client
        self.sensordict = sensordict 
        self.confdict = confdict
        # variables for broadcasting via mqtt:
        self.count=0
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        ###
        self.port = confdict['serialport']+sensordict.get('port')
        self.baudrate = int(sensordict.get('baudrate'))
        if not self.baudrate == 115200:
            log.msg('baudrate for Fluke 289 must be 115200. exiting...')
        self.port = confdict['serialport']+sensordict.get('port')
        self.parity=sensordict.get('parity')
        self.bytesize=sensordict.get('bytesize')
        self.stopbits=sensordict.get('stopbits')
        self.timeout=0 # should be rate depended
    
        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)

        # switch on debug mode
        debugtest = confdict.get('debug')
        self.debug = False
        if debugtest == 'True':
            log.msg('DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))
        # query serial number at startup
        self.get_serialnumber = False
        # serial number of config file will be checked at startup
        self.sn = '' # self.sensordict['serialnumber']
        self.oldsensorid = ''
        self.ser = None



    def open_connection(self):
        log.msg("connecting to serial port")
        try:
            ser = serial.Serial(self.port, baudrate=int(self.baudrate), parity=self.parity, bytesize=int(self.bytesize), stopbits=int(self.stopbits), timeout=self.timeout)
        except:
            return False
        self.ser = ser
        return ser

    def write_read(self,ser,command,end="MARTASEND",maxcnt=20,debug=False):
        data = b""
        if ser:
            # don't produce log entries, when there isn't even a connection
            try:
                ser.flush()
                if sys.version_info >= (3, 0):
                    ser.write(command.encode("utf-8"))
                else:
                    ser.write(command)
                time.sleep(0.05)
                data=ser.readline()
                if sys.version_info >= (3, 0):
                    data = str(data.decode("utf-8"))
                else:
                    data = str(data)
                if not data:
                    if debug:
                        log.msg("DEBUG - got empty line")
                datalist = data.split('\r')
                return datalist
            except serial.SerialException as e:
                if self.debug:
                    log.msg("SerialException found ({})".format(e))
                time.sleep(1)
                #self.restart()
            except:
                log.msg("Other exception found")
                raise

    def sendRequest(self):
        # connect to serial
        ser = self.ser
        if not ser or not ser.isOpen():
            ser = self.open_connection()
            if ser:
                self.get_serialnumber = True
                return
        if self.get_serialnumber:
            answer = self.write_read(self.ser,'ID\r\n',debug=self.debug)
            if not len(answer) == 3:
                if self.debug:
                    log.msg('error: got invalid serial number: '+str(answer))
                return
            error = answer[0]
            datastring = answer[1]
            data = datastring.split(',')
            model = data[0]
            sw_version = data[1]
            sn = data[2]
            log.msg(str(error)+'\t'+model+'\t'+sw_version+'\t'+sn)
            if not sn == self.sensordict['serialnumber']:
                log.msg('S/N '+sn+' and the one given in the config file '+self.sensordict['serialnumber']+' differ')
            self.sn=sn
            self.get_serialnumber = False
            return

        if not ser:
            # exit, when there is no serial connection
            return
        answer = self.write_read(ser,'QM\r\n',debug=self.debug)
        if not len(answer) == 3:
            if self.debug:
                log.msg('error: got invalid data: '+str(answer))
            return
        error = answer[0]
        datastring = answer[1]
        data = datastring.split(',')
        value = data[0]
        unit = data[1]
        state = data[2]
        state_nr = STATE.index(state)
        attribute = data[3]
        attribute_nr = ATTRIBUTE.index(attribute)
        if self.debug:
            log.msg(str(error)+'\t'+value+'\t'+unit+'\t'+state+'\t'+str(state_nr)+'\t'+attribute+'\t'+str(attribute_nr))

        t = datetime.utcnow()
        darray = datetime2array(t)
        packcode = "<6hLl"
        # header
        packcodeH = "6hLl"
        valuelist = value.split('E')
        if valuelist[1] == '+37':
            # invalid data
            if self.debug:
                log.msg('invalid data (+9.9999999E+37)')
            return
        prefix = PREFIX[int(valuelist[1])//3+3]
        valueint = int(float(valuelist[0]) * 10000)
        if self.debug:
            log.msg(str(valueint)+'\t'+unit)
        darray.append(valueint)
        instrument = self.sensordict['name']+prefix+unit
        unit = '['+prefix+unit+']'
        # redefine sensorid e.g. FLUKE289VAC_123456_0001
        sensorid = self.sensordict['sensorid']
        sensorid = instrument+'_'+self.sn+'_'+sensorid.split('_')[2]
        if not sensorid == self.oldsensorid:
            self.oldsensorid = sensorid
            log.msg('switched to '+sensorid)
        instrument = '['+instrument+']'
        header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid,'[var1]',instrument,unit,'[10000]',packcodeH,struct.calcsize(packcode))
        data_bin = struct.pack(packcode,*darray)
        # date of dataloggers timestamp
        filedate = datetime.strftime(datetime(darray[0],darray[1],darray[2]), "%Y-%m-%d")
        if not self.confdict.get('bufferdirectory','') == '':
            #acs.dataToFile(self.confdict.get('bufferdirectory'), sensorid, filedate, data_bin, header)
            mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filedate, data_bin, header)
            if self.debug:
                log.msg('Daten gesichert...')

        # sending via MQTT
        data = ','.join(list(map(str,darray)))
        senddata = False
        head = header
        #topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
        topic = self.confdict.get('station') + '/' + sensorid
        coll = int(self.sensordict.get('stack'))
        if coll > 1:
            self.metacnt = 1 # send meta data with every block
            if self.datacnt < coll:
                self.datalst.append(data)
                self.datacnt += 1
            else:
                senddata = True
                data = ';'.join(self.datalst)
                self.datalst = []
                self.datacnt = 0
        else:
            senddata = True

        if senddata:
            self.client.publish(topic+"/data", data)
            if self.count == 0:
                self.client.publish(topic+"/meta", head)
            self.count += 1
            if self.count >= self.metacnt:
                self.count = 0





# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
import sys
import numpy as np
from datetime import datetime, timezone
from twisted.protocols.basic import LineReceiver
from twisted.python import log
from martas.core import methods as mm


## GEM -GSM19 protocol
##
class GSM19Protocol(LineReceiver):
    """
    The GSM 19 protocol for extracting RTT data from GSM19
    SETUP of GSM19:
        1.) in the main menu go to C-Info
        2.) in info select B-RS232
        3.) choose BAUD rate of 115200 - press F
        4.) real time RS232 transmission: select yes and press F
        5.) F again, thats it go back to main menu
    Supported modes:
        base

    Treatment of GPS and NTP time:
    The protocol determines two times: 1) GPS time provided by the data logger
    2) NTP time of the PC when data is received by serial comm.
    Primary time is selected in sensor.cfg (either NTP or GPS)
    Timediff between GPS and NTP is recorded, and a 1000 line median average is send
    within the DataDictionary (DataNTPTimeDelay).
    Besides, a waring message is generated if NTP and GPS times differ by more than 3 seconds.
    For this comparison (and only here) a typical correction (delay) value is applied to NTP.
    e.g. 6.2 sec for POS1.
    """

    def __init__(self, client, sensordict, confdict):
        """
        'client' could be used to switch between different publishing protocols
                 (e.g. MQTT or WS-MCU gateway factory) to publish events
        'sensordict' contains a dictionary with all sensor relevant data (sensors.cfg)
        'confdict' contains a dictionary with general configuration parameters (martas.cfg)
        """
        print ("  -> Initializing GSM19...")
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        #log.msg("  -> Sensor: {}".format(self.sensor))
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        self.errorcnt = {'time':0}
        self.timesource = self.sensordict.get('ptime','')
        self.delaylist = []  # delaylist contains up to 1000 diffs between gps and ntp
                             # the median of this values is used for ntp timedelay
        self.timedelay = 0.0
        self.timethreshold = 3 # secs - waring if timedifference is larger the 3 seconds

        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)
        # PYTHON version
        self.pvers = sys.version_info[0]

    def connectionMade(self):
        log.msg('  -> {} connected.'.format(self.sensor))

    def connectionLost(self, reason):
        log.msg('  -> {} lost.'.format(self.sensor))

    def processData(self, data):

        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        date = datetime.strftime(currenttime, "%Y-%m-%d")
        actualtime = datetime.strftime(currenttime, "%Y-%m-%dT%H:%M:%S.%f")
        outtime = datetime.strftime(currenttime, "%H:%M:%S")
        filename = date
        timestamp = datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")
        intensity = 88888.8
        typ = "none"
        dontsavedata = False

        packcode = '6hLLl6hL'
        header = "# MagPyBin %s %s %s %s %s %s %d" % (self.sensor, '[f,var1,sectime]', '[f,err,sectime]', '[nT,none,none]', '[1000,1000,1]', packcode, struct.calcsize('<'+packcode))

        try:
            # Extract data
            data_array = data
            if len(data_array) == 2:
                typ = "oldbase"
            elif len(data_array) == 3:
                typ = "valid"
            # add other types here
        except:
            log.err('GSM19 - Protocol: Output format not supported - use either base, ... or mobile')

        # Extracting the data from the station
        # Extrat time info and use as primary if GPS is on (in this case PC time is secondary)
        #                          PC is primary when a GPS is not connected

        if typ == "valid" or typ == "oldbase": # Comprises Mobile and Base Station mode with single sensor and no GPS
            intensity = float(data_array[1])
            try:
                systemtime = datetime.strptime(date+"-"+data_array[0], "%Y-%m-%d-%H%M%S.%f")
            except:
                # This exception happens for old GSM19 because time is
                # provided e.g. as 410356 instead of 170356 for 17:03:56 (Thursday)
                # e.g 570301.0 instead of 09:03:01 (Friday)
                try:
                    hournum = int(data_array[0][:-6])
                    rest = data_array[0][-6:]
                    factor = np.floor(hournum/24.) # factor = days since starting
                    hour = int(hournum - factor*24.)
                    systemtime = datetime.strptime(date+"-"+str(hour)+rest, "%Y-%m-%d-%H%M%S.%f")
                    #print ("Got oldbase systemtime")
                except:
                    systemtime = currenttime
                    self.timesource = 'NTP'
            if len(data_array) == 2:
                typ = "base"
                errorcode = 99
            elif len(data_array[2]) == 3:
                typ = "base"
                errorcode = int(data_array[2])
            else:
                typ = "gradient"
                gradient = float(data_array[2])
        elif typ == "none":
            dontsavedata = True
            pass

        gpstime = datetime.strftime(systemtime, "%Y-%m-%d %H:%M:%S.%f")

        try:
            # Analyze time difference between GSM internal time and utc from PC
            timelist = sorted([systemtime,currenttime])
            timediff = timelist[1]-timelist[0]
            #secdiff = timediff.seconds + timediff.microseconds/1E6
            delta = timediff.total_seconds()
            if not delta in [0.0, np.nan, None]:
                self.delaylist.append(timediff.total_seconds())
                self.delaylist = self.delaylist[-1000:]
            if len(self.delaylist) > 100:
                try:
                    self.timedelay = np.median(np.asarray(self.delaylist))
                except:
                    self.timedelay = 0.0
            if delta > self.timethreshold:
                self.errorcnt['time'] +=1
                if self.errorcnt.get('time') < 2:
                    log.msg("{} protocol: large time difference observed for {}: {} sec".format(self.sensordict.get('protocol'), self.sensor, delta))
            else:
                self.errorcnt['time'] = 0
        except:
            pass

        if self.sensordict.get('ptime','') in ['NTP','ntp']:
            secondtime = gpstime
            maintime = timestamp
        else:
            maintime = gpstime
            secondtime = timestamp

        try:
            if not typ == "none":
                # extract time data
                datearray = mm.time_to_array(maintime)
                try:
                    datearray.append(int(intensity*1000.))
                    if typ == 'base':
                        datearray.append(int(errorcode*1000.))
                    else:
                        datearray.append(int(gradient*1000.))
                    internalarray = mm.time_to_array(secondtime)
                    datearray.extend(internalarray)
                    data_bin = struct.pack('<'+packcode,*datearray)
                except:
                    log.msg('GSM19 - Protocol: Error while packing binary data')
                    pass
        except:
            log.msg('GSM19 - Protocol: Error with binary save routine')
            pass

        if not self.confdict.get('bufferdirectory','') == '':
            mm.data_to_file(self.confdict.get('bufferdirectory'), self.sensor, filename, data_bin, header)

        return ','.join(list(map(str,datearray))), header


    def lineReceived(self, line):
        topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
        try:
            if self.pvers > 2:
                line=line.decode('ascii')
            # extract only ascii characters
            line = ''.join(filter(lambda x: x in string.printable, str(line)))
        except:
            log.msg ("gsm19protocol: error while decoding received line: {}".format(line))
            line = ''

        ok = True
        try:
            data = line.split()
            if len(data) >= 2:
                data, head = self.processData(data)
            else:
                log.msg('{}: Data seems not to be appropriate G19 data. Received data looks like: {}'.format(self.sensordict.get('protocol'),line))
                ok = False

            if ok:
                senddata = False
                coll = int(self.sensordict.get('stack'))
                if coll > 1:
                    self.metacnt = 1 # send meta data with every block
                    if self.datacnt < coll:
                        self.datalst.append(data)
                        self.datacnt += 1
                    else:
                        senddata = True
                        data = ';'.join(self.datalst)
                        self.datalst = []
                        self.datacnt = 0
                else:
                    senddata = True

                if senddata:
                    self.client.publish(topic+"/data", data, qos=self.qos)
                    if self.count == 0:
                        add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDecription:{},DataTimeProtocol:{},DataNTPTimeDelay:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''),self.timesource, self.timedelay )
                        self.client.publish(topic+"/dict", add, qos=self.qos)
                        self.client.publish(topic+"/meta", head, qos=self.qos)
                    self.count += 1
                    if self.count >= self.metacnt:
                        self.count = 0
        except:
            log.err('{}: Unable to parse data {}'.format(self.sensordict.get('protocol'), line))



#!/usr/bin/env python
"""
DESCRIPTION
    Checking a MagPy database structure for tables which are not contained in DATAINFO and SENSORS.
    The background/necessity of this method is related to archive.py and db_truncate.py, which make
    use of the DATAINFO table for selecting data tables to be archived AND cleaned. Therefore, tables
    not listed in DATAINFO are growing endless until the memory is full.
    
METHOD
    Selects all tables which end with with typical revision numbers (i.e 0001) and check for their
    existance in DATAINFO and SENSORS. 

GROUP
   MARTAS app 

REQUIREMENTS:
   MagPy >= 1.0.0
   Database credentials stored with addcred
"""
# Define packages to be used (local refers to test environment)
# ------------------------------------------------------------

from magpy.core import database
from martas.core import methods as mm

import getopt
import sys
import socket


def add_datainfo(db, tableonly=None, verbose=False, debug=False):
    """
    DESCRIPTION
    """
    if not tableonly:
        tableonly = []
    if debug:
        print (" Adding tableonly to DATAINFO...")
    
    if not len(tableonly) > 0:
        return False
    # get a list with all datainfoids covering the selected time range
    cursor = db.cursor()

    for tab in tableonly:
        sql = 'INSERT INTO DATAINFO (DataID) VALUES ("{}")'.format(tab)
        if not debug:
            try:
                cursor.execute(sql)
            except:
                print ("   Error when sending sql query")
        else:
            print (" if not debug I would execute: {}".format(sql))
    cursor.close()
    return True


def obtain_datainfo(db, blacklist=None, verbose=False, debug=False):

    if not blacklist:
        blacklist = []
    resultdict = {}
    addstr = ''
    # looks like {'data_1_0001_0001' : {'mintime':xxx, 'maxtime' : xxx }, ...}

    if debug:
         print ("   Creating sql query...")
    # get a list with all datainfoids covering the selected time range
    sql = 'SELECT DataID,DataMinTime,DataMaxTime FROM DATAINFO'

    if len(blacklist) > 0:
        sql = "{} WHERE".format(sql)
        addstr = ''
        for el in blacklist:
            addstr += " AND DataID NOT LIKE '{}%'".format(el)
        addstr = addstr.replace(" AND","",1)
    sql = "{}{}".format(sql,addstr)
    if debug:
         print ("   -> query looks like: {}".format(sql))

    cursor = db.db.cursor()
    message = db._executesql(cursor, sql)
    if message:
        print(message)
    datainfolist =  cursor.fetchall()
    for el in datainfolist:
        # check whether a data table with this name is existing
        verifytable = "SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = '{}'".format(el[0])
        cursor.execute(verifytable)
        test =  cursor.fetchall()
        if test:
            resultdict[el[0]] = {'exists':'BOTH', 'mintime': el[1], 'maxtime': el[2]}
        else:
            resultdict[el[0]] = {'exists':'DATAINFO', 'mintime': el[1], 'maxtime': el[2]}
    cursor.close()
    if debug:
        print ("   -> Obtained the following DataIDs:", resultdict)

    return resultdict


def match(first, second):
    """
    DESCRIPTION
        search string with wildcards in other string
        corrected  in len(first) >= 1
    REFERENCE
        https://www.geeksforgeeks.org/wildcard-character-matching/
    """
 
    # If we reach at the end of both strings, we are done
    if len(first) == 0 and len(second) == 0:
        return True
 
    # Make sure that the characters after '*' are present
    # in second string. This function assumes that the first
    # string will not contain two consecutive '*'
    if len(first) > 1 and first[0] == '*' and  len(second) == 0:
        return False
 
    # If the first string contains '?', or current characters
    # of both strings match
    if (len(first) >= 1 and first[0] == '?') or (len(first) != 0
        and len(second) !=0 and first[0] == second[0]):
        return match(first[1:],second[1:])
 
    # If there is *, then there are two possibilities
    # a) We consider current character of second string
    # b) We ignore current character of second string.
    if len(first) !=0 and first[0] == '*':
        return match(first[1:],second) or match(first,second[1:])
 
    return False
 

def get_tables(db,identifier="*_00??",verbose=False,debug=False):
    """
    DESCRIPTION
        get all tables which end with the identifier
    RETURN
        list with table names
        (asume table name == DataID, and table name-identifier = SensorId)        
    """
    if verbose or debug:
        print (" Getting all tables ...")
    cursor = db.db.cursor()

    if debug:
        print (" - Current database information:")
        db.info(destination='stdout',level='full')

    tablessql = 'SHOW TABLES'
    message = db._executesql(cursor, tablessql)
    if message:
        print(message)
    tables = cursor.fetchall()
    cursor.close()
    tables = [el[0] for el in tables]

    if debug:
        print (" Found: {}".format(tables))

    if not len(tables) > 0:
        print (' checkdatainfo.py: no tables found in specified database - aborting')
        sys.exit()

        
    # Match tables with identifier
    idtabs = [el for el in tables if match(identifier, el)]
    if debug:
        print (" Found after matching with ID {}: {}".format(identifier, idtabs))

    return idtabs


def matchtest(identifier,string):
    if match(identifier,string):
        print ("Yes")
    else:
        print ("No") 

#matchtest('*_00?1','WIC_adjusted_0001_0001')
#matchtest('*_00*','WIC_adjusted_0001_0001')
#matchtest('*_00??','WIC_adjusted_0001_0001')
#matchtest('*_0001','WIC_adjusted_0001_0001')
#matchtest('*_02??','WIC_adjusted_0001_0001')


def main(argv):
    version = "2.0.0"
    cred = ''
    identifier = ''
    blacklist = []
    checkdatainfo = False
    checksensors = False
    add = False
    verbose = False
    hostname = socket.gethostname().upper()
    debug=False
    
    head = 'checkdatainfo.py -c <cred> -i <id> -d <datainfo> -s <sensors> -a <add>'
    try:
        opts, args = getopt.getopt(argv,"hc:i:dsaD",["credentials=","id=","datainfo=","sensors=","add=","debug=",])
    except getopt.GetoptError:
        print (head)
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print ('-------------------------------------')
            print ('Description:')
            print ('-- checkdatainfo.py checks for missing tables in DATAINFO  --')
            print ('-- and SENOSRS.           --')
            print ('-------------------------------------')
            print ('Usage:')
            print (head)
            print ('-------------------------------------')
            print ('Options:')
            print ('-c (required) : credentials for a database')
            print ('-i            : data table identifiers - end of table name i.e "00??" (? can be numbers from 0-9)')
            print ('-d            : check datainfo')
            print ('-s            : check sensors')
            print ('-a            : add missing data to DATAINFO ( if "-d") and SENSORS (if "-s")')
            print ('-------------------------------------')
            print ('Example:')
            sys.exit()
        elif opt in ("-c", "--credentials"):
            cred = arg
        elif opt in ("-i", "--id"):
            identifier = arg
        elif opt in ("-d", "--datainfo"):
            checkdatainfo = True
        elif opt in ("-s", "--sensors"):
            checksensors = True
        elif opt == "-a":
            add = True
        elif opt == "-v":
            verbose = True
        elif opt in ("-D", "--debug"):
            debug = True

    if verbose or debug:
        print ("Running checkdatainfo.py version {}".format(version))
        print ("-------------------------------")

    if not checkdatainfo and not checksensors:
        print ('Specify either -d and/or -s')
        print ('-- Check checkdatainfo.py -h for more options and requirements')
        sys.exit()

    if cred == '':
        print ('Specify database credentials using the  -c option:')
        print ('-- Database credentials are created by addcred')
        print ('-- Check checkdatainfo.py -h for more options and requirements')
        sys.exit()


    # 1. Connect to database
    db = mm.connect_db(cred)

    # 2. Check tables
    tables = get_tables(db,identifier=identifier,verbose=verbose,debug=debug)

    # 3. Check for tables in DATAINFO
    datainfodict = obtain_datainfo(db, blacklist=blacklist, verbose=verbose, debug=debug)
    tableonly = [tab for tab in tables if not tab in datainfodict]
    for tab in tableonly:
        datainfodict[tab]  = {'exists':'TABLE'}
    if verbose or debug:
        for el in datainfodict:
            dd = datainfodict.get(el)
            if dd.get('exists') == 'BOTH':
                print ("Existing in BOTH:  {}".format(el) )
            elif dd.get('exists') == 'TABLE':
                print ("Existing only as TABLE:  {}".format(el) )
            elif dd.get('exists') == 'DATAINFO':
                print ("Existing only in DATAINFO:  {}".format(el) )
    if add:
        print ("Adding DATAINFO information")
        if debug:
            print ("Table only", tableonly) 
        #add_datainfo(db,tableonly=tableonly, verbose=verbose, debug=debug)
        print ("Better choice.. add the missing tables in xMagPy and export the header to the database")
        print (" routinely run this method (at least once per month and report tables")

    # 4. Check for tables in SENSORS


    if verbose or debug:
        print ("----------------------------------------------------------------")
        print ("checking datainfo app finished")
        print ("----------------------------------------------------------------")
        print ("SUCCESS")

if __name__ == "__main__":
   main(sys.argv[1:])






# ###################################################################
# Import packages
# ###################################################################

import sys    # for version identification
import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
from datetime import datetime, timezone
from twisted.python import log
from martas.core import methods as mm

try:
    import pyownet
    onewire = True
    owsensorlist = []
except:
    print("Onewire (pyownet) package not available")
    onewire = False


typedef = {'DS18B20':['temperature'], 'DS2438':['temperature','VAD','VDD','humidity','vis'], 'DS1420':['well'], 'DS18S20':['temperature']}


if onewire:
    class OwProtocol():
        """
        Protocol to read one wire data from usb DS unit.
        For this protocol "owserver" needs to be running.
        All connected sensors are listed and data is distributed in dependency of sensor id
        Dipatch url links are defined by channel 'ow' and id+'value'
        Known issues:
        When renaming th PC, owserver can not properly be reached. -> Restart 
        """
        # TODO check humidity and pressure
        def __init__(self, client, sensordict, confdict):
            log.msg("  -> one wire: Initializing ...")
            self.client = client
            self.sensordict = sensordict
            self.confdict = confdict
            self.owhost = confdict.get('owhost')
            self.owport = int(confdict.get('owport'))
            self.sensor = sensordict.get('sensorid') # should be Ow
            self.hostname = socket.gethostname()
            self.printable = set(string.printable)
            self.reconnectcount = 0
            self.removelist = [] # list of sensorspaths from sensors.cfg which are not found
            # Extract eventually existing one wire sensors from sensors.cfg
            log.msg("  -> one wire: Checking existing sensors ...")
            self.existinglist = mm.get_sensors(confdict.get('sensorsconf'),identifier='!')
            log.msg("  -> one wire: Checking for new sensors ...")
            self.sensorarray = self.GetOneWireSensorList(self.existinglist)
            self.count = [0]*len(self.sensorarray)  ## counter for sending header information
            self.metacnt = 2  # Send header information often for OW
            self.datalst = [[]]*len(self.sensorarray)
            self.datacnt = [0]*len(self.sensorarray)
            log.msg("  -> one wire: Initialized")
            #print (self.existinglist)
            #print (self.count)

            # debug mode
            debugtest = confdict.get('debug')
            self.debug = False
            if debugtest == 'True':
                log.msg('     DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
                self.debug = True    # prints many test messages
            else:
                log.msg('  -> Debug mode = {}'.format(debugtest))

            # QOS
            self.qos=int(confdict.get('mqttqos',0))
            if not self.qos in [0,1,2]:
                self.qos = 0
            log.msg("  -> setting QOS:", self.qos)


        def GetOneWireSensorList(self, existinglist=None):
            if not existinglist:
                existinglist = []
            try:
                self.owproxy = pyownet.protocol.proxy(host=self.owhost, port=self.owport)
                sensorlst = self.owproxy.dir()
            except:
                log.msg("  -> one wire: could not contact to owhost")
                return []
            #log.msg("  -> one wire: {}".format(sensorlst))
            # Python3 checks
            #if sys.version_info >= (3, 0):
            # compare currently read sensorlst with original sensorlst (eventually from file)
            existingpathlist = [line.get('path') for line in existinglist]
            # Identify attached sensors and their types
            idlist = []
            fakelst = []
            # check for sensors which should be there but are not
            notfound = [el for el in existingpathlist if not el in sensorlst]
            for el in notfound:
                if not el in self.removelist:
                    log.msg("OW: sensor with path {} (as listed in sensors.cfg) is not found".format(el))
                    self.removelist.append(el)

            for el in sensorlst:
                values = {}
                # get posistions of sensorid, protocol(OW), name(DS18B20), serialnumber, revision and path
                # ADD SENSORDICT ELEMENTS to acquisition support
                #'sensorid','port','baudrate','bytesize','stopbits', 'parity','mode','init','rate',       				'protocol','name','serialnumber','revision','path','pierid','sensordesc'
                if not el in fakelst:
                    # make an array of IDs
                    line = []
                    idel = el.replace('/','').replace('.','').strip()
                    line.append(idel)
                    line.append(el)
                    #line.append(sensorname)
                    # make a dict for each ID
                    path = el+'type'
                    typ = self.owproxy.read(path)
                    #log.msg(typ)
                    # Python3 checks
                    if sys.version_info >= (3, 0):
                       try:
                           typ = typ.decode('ascii')
                       except:
                           pass
                    if not typ == 'DS1420': # do not add dongle
                        if el in self.removelist:
                            # el found again
                            # Dropping it from removelist
                            log.msg("OW: sensor with path {} is active again".format(el)) 
                            self.removelist.remove(el)
                        if el in existingpathlist:
                            #print ("Sensor {} existing".format(path))
                            values = [line for line in existinglist if line.get('path') == el][0]
                        else:
                            values['path'] = el
                            values['serialnumber'] = idel
                            values['name'] = typ
                            values['protocol'] = 'Ow'
                            revision = '0001'
                            values['revision'] = revision
                            values['stack'] = 0
                            values['sensorid'] = '{}_{}_{}'.format(typ,idel,revision)
                            log.msg("OW: Writing new sensor input to sensors.cfg ...")
                            success = mm.add_sensors(self.confdict.get('sensorsconf'), values, block='OW')
                            #success = mm.add_sensors(self.confdict.get('sensorsconf'), values, block='OW')
                            if success:
                                log.msg("    {} written".format(values.get('sensorid')))
                                # extend existingpathlist
                                self.existinglist.append(values)
                        idlist.append(values)
            return idlist

        #def connectionMade(self):
        #    log.msg('%s connected.' % self.sensor)

        #def connectionLost(self, reason):
        #    log.msg('%s lost.' % self.sensor)


        def sendRequest(self):
            #log.msg("Sending periodic request ...")
            sensorarray = self.GetOneWireSensorList(self.existinglist)
            if not len(self.count) == len(sensorarray):
                # if length of sensorarray is changing - reset counters 
                self.count = [0]*len(sensorarray)
                self.datalst = [[]]*len(self.sensorarray)
                self.datacnt = [0]*len(self.sensorarray)
            for idx, line in enumerate(sensorarray):
                if self.debug:
                    print ("Getting sensor ID:", line.get('sensorid'))
                sensorid = line.get('sensorid')
                valuedict = {}
                if self.debug:
                    print ("Available parameters:")
                    print (self.owproxy.dir(line.get('path')))
                if typedef.get(line.get('name')):  # check if sensor exists
                    for para in typedef.get(line.get('name'),[]):
                        path = line.get('path')+para
                        if para == 'humidity': ## Check whether separete treatment is necessary
                            pass
                            #line.get('path')+para
                            #print ("ALL", self.owproxy.dir())
                            #print ("sens",self.owproxy.dir(line.get('path')))
                            #if para == 'pressure':
                            #    #if sensortypus == "pressure":
                            #    #    #print "Pressure [hPa]: ", self.mpxa4100(vad,temp)
                            #    #    humidity = self.mpxa4100(vad,temp)
                            #    pass
                        valuedict[para] = self.owproxy.read(path)

                    topic = self.confdict.get('station') + '/' + sensorid
                    data, head  = self.processOwData(sensorid, valuedict)

                    # To find out, which parameters are available use:
                    #print (self.owproxy.dir(line.get('path')))

                    senddata = False
                    try:
                        coll = int(line.get('stack'))
                    except:
                        coll = 0
                    #coll = int(self.sensordict.get('stack'))
                    if coll > 1:
                        self.metacnt = 1 # send meta data with every block for stacked transfer
                        if self.datacnt[idx] < coll:
                            self.datalst[idx].append(data)
                            self.datacnt[idx] += 1
                        else:
                            senddata = True
                            data = ';'.join(self.datalst[idx])
                            self.datalst[idx] = []
                            self.datacnt[idx] = 0
                    else:
                        senddata = True

                    if senddata:
                        self.client.publish(topic+"/data", data, qos=self.qos)
                        if self.count[idx] == 0:
                            ## 'Add' is a string containing dict info like: 
                            ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,... 
                            add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDecription:{},DataTimeProtocol:{}".format( line.get('sensorid',''),self.confdict.get('station',''),line.get('pierid',''),line.get('protocol',''),line.get('sensorgroup',''),line.get('sensordesc',''),line.get('ptime','') )
                            #print ("...", add)
                            self.client.publish(topic+"/dict", add, qos=self.qos)
                            self.client.publish(topic+"/meta", head, qos=self.qos)
                            self.count[idx] += 1
                        if self.count[idx] >= self.metacnt:
                            self.count[idx] = 0


        def processOwData(self, sensorid, datadict):
            """Process OW data """
            key, ele, unit = '','',''
            currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
            outdate = datetime.strftime(currenttime, "%Y-%m-%d")
            filename = outdate
            actualtime = datetime.strftime(currenttime, "%Y-%m-%dT%H:%M:%S.%f")
            outtime = datetime.strftime(currenttime, "%H:%M:%S")
            timestamp = datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")
            packcode = '6hL'+'l'*len(datadict)
            multplier = str([1000]*len(datadict)).replace(' ','')
            if sensorid.startswith('DS18'):
                key = '[t1]'
                ele = '[T]'
                unit = '[degC]'
            elif sensorid.startswith('DS2438'):
                #'temperature','VAD','VDD','humidity','vis'
                key = '[t1,var1,var2,var3,var4]'
                ele = '[T,RH,VDD,VAD,VIS]'
                unit = '[degC,per,V,V,V,V]'

            header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid, key, ele, unit, multplier, packcode, struct.calcsize('<'+packcode))

            data_bin = None
            datearray = ''
            try:
                datearray = mm.time_to_array(timestamp)
                paralst = typedef.get(sensorid.split('_')[0])
                for para in paralst:
                    if para in datadict:
                        datearray.append(int(float(datadict[para])*1000))
                data_bin = struct.pack('<'+packcode,*datearray)  # little endian
            except:
                log.msg('Error while packing binary data')

            if not self.confdict.get('bufferdirectory','') == '' and data_bin:
                mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)
            #print ("Sending", ','.join(list(map(str,datearray))), header)
            return ','.join(list(map(str,datearray))), header




#!/usr/bin/env python
# coding=utf-8

"""
Testing module for methods from apps
"""

import unittest

import sys
sys.path.insert(1,'/home/leon/Software/MARTAS/') # should be magpy2
sys.path.insert(1,'/home/leon/Software/magpy/') # should be magpy2

import os
import shutil
import numpy as np
from magpy.stream import DataStream, read
from magpy.core import database
from magpy.opt import cred as mpcred

from datetime import datetime, timedelta, timezone
from martas.core import methods as mameth

from martas.app import archive
from martas.app import basevalue
from martas.app import checkdatainfo
from martas.app import db_truncate
from martas.app import file_download
from martas.app import filter
from martas.app import threshold


def create_teststream(startdate=datetime(2022, 11, 22)):
    teststream = DataStream()
    array = [[] for el in DataStream().KEYLIST]
    array[1] = [20000] * 720
    array[1].extend([22000] * 720)
    array[1] = np.asarray(array[1])
    array[2] = np.asarray([0] * 1440)
    array[3] = np.asarray([20000] * 1440)
    array[6] = np.asarray([np.nan] * 1440)
    # array[4] = np.sqrt((x*x) + (y*y) + (z*z))
    array[0] = np.asarray([startdate + timedelta(minutes=i) for i in range(0, len(array[1]))])
    rain1 = np.arange(1,721,1.0)
    rain2 = np.arange(1,361,0.5)
    rain = np.concatenate((rain1, rain2))
    array[7] = rain
    array[DataStream().KEYLIST.index('sectime')] = np.asarray(
        [startdate + timedelta(minutes=i) for i in range(0, len(array[1]))]) + timedelta(minutes=15)
    teststream = DataStream(header={'SensorID': 'Test_0002_0001'}, ndarray=np.asarray(array, dtype=object))
    teststream.header['col-x'] = 'X'
    teststream.header['col-y'] = 'Y'
    teststream.header['col-z'] = 'Z'
    teststream.header['col-t2'] = 'Text'
    teststream.header['col-var1'] = 'Rain'
    teststream.header['unit-col-x'] = 'nT'
    teststream.header['unit-col-y'] = 'nT'
    teststream.header['unit-col-z'] = 'nT'
    teststream.header['unit-col-t2'] = 'degC'
    teststream.header['unit-col-var1'] = 'mm'
    teststream.header['DataComponents'] = 'XYZ'
    return teststream

class TestArchive(unittest.TestCase):
    """
    Test environment for all methods
    """

    def test_create_datelist(self):
        dl = archive.create_datelist(startdate='', depth=10, debug=True)
        self.assertEqual(len(dl), 10)

    #def test_create_data_selectionlist(self):
    #    dt = archive.create_data_selectionlist(blacklist=None, debug=False)
    #    self.assertEqual(ar[1],11)

    #def test_get_data_dictionary(self):
    #    cfg = archive.get_data_dictionary(db,sql,debug=False)
    #    self.assertEqual(cfg.get("station"),"myhome")

    #def test_get_parameter(self):
    #    sens = archive.get_parameter(plist, debug=False)
    #    self.assertEqual(sens,[])

    #def test_validtimerange(self):
    #    archive.validtimerange(timetuple, mintime, maxtime, debug=False)

class TestBasevalue(unittest.TestCase):
    """
    Test environment for all basevalue methods
    """

    def test_check_conf(self):
        config = mameth.get_conf(os.path.join('..', 'conf', 'basevalue.cfg'))
        startdate, enddate = None, None
        varios, scalars, piers = [],[],[]
        config = basevalue.check_conf(config, startdate, enddate, varios=varios, scalars=scalars, piers=piers, debug=True)
        pl = config.get('pierlist')
        self.assertEqual(len(pl), 2)

    """
    def test_get_runmode(self):
        joblist = ['firstrun','overview']
        jl, jt = basevalue.get_runmode(joblist)
        self.assertEqual(jt, 'firstrun')

    def test_prepare_and_backup_file(self):
        testpath = os.path.join('/tmp', 'BLV_LEMI036_1_0002_GP20S3NSS2_012201_0001_A2.txt')
        print ("Copying BLV to temporary directory")
        shutil.copyfile(os.path.join('..', 'test', 'BLV_LEMI036_1_0002_GP20S3NSS2_012201_0001_A2.txt'),testpath)
        np = basevalue.backup_file(testpath)
        self.assertTrue(np)
        starttime = "2024-01-01"
        endtime = "2024-12-31"
        data = read(testpath)
        lb = len(data)
        print ("Trimming BLV in temporary directory")
        basevalue.prepare_file(testpath, starttime, endtime, format_type='PYSTR')
        data = read(testpath)
        la = len(data)
        self.assertEqual(lb-la, 119)

    def test_prepare_table(self):
        starttime = "2024-01-01"
        endtime = "2024-12-31"
        testpath = os.path.join('/tmp', 'BLV_LEMI036_1_0002_GP20S3NSS2_012201_0001_A2.txt')
        credentials = 'cobsdb'
        db = database.DataBank(host=mpcred.lc(credentials, 'host'), user=mpcred.lc(credentials, 'user'),
                                   password=mpcred.lc(credentials, 'passwd'), database=mpcred.lc(credentials, 'db'))
        shutil.copyfile(os.path.join('..', 'test', 'BLV_LEMI036_1_0002_GP20S3NSS2_012201_0001_A2.txt'),testpath)
        data = read(testpath)
        lb = len(data)
        db.write(data, tablename = 'BLV_LEMI036_1_0002_GP20S3NSS2_012201_0001_A2', mode = 'replace')
        dl = basevalue.prepare_table(db, 'BLV_LEMI036_1_0002_GP20S3NSS2_012201_0001_A2', starttime, endtime, tcol='time', cond=None)
        d = db.read('BLV_LEMI036_1_0002_GP20S3NSS2_012201_0001_A2')
        la = len(d)
        self.assertEqual(lb-la, 119)

    def test_baseline_overview(self):
        config = mameth.get_conf(os.path.join('..', 'conf', 'basevalue.cfg'))
        config['dbcredentials'] = "cobsdb"
        startdate, enddate = None, "2023-12-31"
        varios, scalars, piers = ["LEMI036_1_0002_0002"],["GP20S3NSS2_012201_0001_0001"],["A2"]
        config = basevalue.check_conf(config, startdate, enddate, varios=varios, scalars=scalars, piers=piers, debug=True)
        config['blvdatapath'] = "/tmp"
        config['blvabb'] = "BLV"
        dl = basevalue.baseline_overview(runmode='firstrun', config=config, debug=True)
        self.assertTrue(dl)

    def test_baseline_recalc(self):
        config = mameth.get_conf(os.path.join('..', 'conf', 'basevalue.cfg'))
        config['dbcredentials'] = "cobsdb"
        startdate, enddate = None, "2025-12-31"
        varios, scalars, piers = ["LEMI036_1_0002_0002"],["GP20S3NSS2_012201_0001_0001"],["A2"]
        config = basevalue.check_conf(config, startdate, enddate, varios=varios, scalars=scalars, piers=piers, debug=True)
        config['blvdatapath'] = "/tmp"
        config['base'] = os.path.abspath("../test")
        config['obscode'] = "WIC"
        config['didatapath'] = os.path.abspath("../test/archive/WIC/DI/analyze")
        config['blvabb'] = "BLV"
        config['writeblv2file'] = True
        print (config)
        runmode = "firstrun"
        basevalue.basevalue_recalc(runmode, config=config, startdate=None, enddate=None, debug=False)
        t = False
        if os.path.isfile(os.path.join("/tmp","BLV_LEMI036_1_0002_GP20S3NSS2_012201_0001_A2_2025_firstrun.txt")):
            t = True
        self.assertTrue(t)
    """


class TestCheckdataInfo(unittest.TestCase):
    """
    Test environment for all methods
    """
    def test_add_datainfo(self):
        #add_datainfo(db, tableonly=None, verbose=False, debug=False)
        pass


    def test_obtain_datainfo(self):
        #xxx = obtain_datainfo(db, blacklist=None, verbose=False, debug=False)
        pass

    def test_match(self):
        #match(first, second)
        pass

    def test_match(self):
        #get_tables(db, identifier="*_00??", verbose=False, debug=False)
        pass

    def test_matchtest(self):
        #matchtest(identifier, string)
        pass


class TestDBTrcuncate(unittest.TestCase):

    def test_query_db(self):
        #query_db(db, sql, debug=False)
        pass

    def test_get_table_tist(self):
        #get_table_tist(db, sensorlist=None, blacklist=None, debug=False)
        pass


class TestFileDownload(unittest.TestCase):
    """
    Test environment for all methods
    """

    def test_check_configuration(self):
        print ("TESTING file_download")
        print ("---------------------------------------------------------------")
        confpath = os.path.abspath(os.path.join(os.path.dirname(__file__),'..','conf','download-source.cfg'))
        config = mameth.get_conf(confpath)
        config, success = file_download.check_configuration(config=config, debug=True)
        #print (config)
        print (success)
        self.assertEqual(config.get('db'),None)
        self.assertEqual(success,True)

    def test_get_datelist(self):
        confpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'conf', 'download-source.cfg'))
        config = mameth.get_conf(confpath)
        config, success = file_download.check_configuration(config=config, debug=False)
        current = datetime.now(timezone.utc).replace(tzinfo=None)
        datelist = file_download.get_datelist(config=config,current=current,debug=True)
        succ = False
        if current.strftime("%Y-%m-%d") in datelist:
            succ = True
        self.assertTrue(succ)

    def test_create_transfer_list_dir_extract(self):
        confpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'conf', 'download-source.cfg'))
        config = mameth.get_conf(confpath)
        config, success = file_download.check_configuration(config=config, debug=False)
        # set protocol to local directory
        config['protocol'] = ''
        current = datetime.now(timezone.utc).replace(tzinfo=None)
        datelist = file_download.get_datelist(config=config,current=current,debug=False)
        filelist = file_download.create_transfer_list(config=config, datelist=datelist, debug=True)
        self.assertTrue(isinstance(filelist, (list,tuple)))

    # missing are file_download.obtain_data_files(config=None,filelist=None,debug=False)
    # and file_download.write_data(config=None,localpathlist=None,debug=False)


class TestFilter(unittest.TestCase):
    """
    Test environment for all methods
    """

    def test_read_conf(self):
        cfg = filter.read_conf("../conf/filter.cfg")
        self.assertNotEqual(cfg,None)

    def test_get_sensors(self):
        recent = True
        dd = filter.read_conf("../conf/filter.cfg")
        groupparameterdict = dd.get('groupparameterdict')
        blacklist = dd.get('blacklist')
        basics = dd.get('basics')
        recentthreshold = int(basics.get('recentthreshold', 7200))
        self.assertEqual(recentthreshold,7200)
        credentials = 'cobsdb'
        db = database.DataBank(host=mpcred.lc(credentials, 'host'), user=mpcred.lc(credentials, 'user'),
                                   password=mpcred.lc(credentials, 'passwd'), database=mpcred.lc(credentials, 'db'))
        highreslst = filter.get_sensors(db=db,groupdict=groupparameterdict,samprate='HF', blacklist=blacklist, recent=recent, recentthreshold=recentthreshold, debug=True)
        print ("Done:", highreslst)

    def test_apply_filter(self):
        sensorlist = []
        basepath = '/srv/archive'
        destination = 'db'
        credentials = 'cobsdb'
        dd = filter.read_conf("../conf/filter.cfg")
        groupparameterdict = dd.get('groupparameterdict')
        blacklist = dd.get('blacklist')
        permanent = dd.get('permanent')
        basics = dd.get('basics')
        outputformat = basics.get('outputformat')
        recentthreshold = int(basics.get('recentthreshold', 7200))
        db = database.DataBank(host=mpcred.lc(credentials, 'host'), user=mpcred.lc(credentials, 'user'),
                                   password=mpcred.lc(credentials, 'passwd'), database=mpcred.lc(credentials, 'db'))
        statusmsg = filter.apply_filter(db, statusmsg={}, groupdict=groupparameterdict, permanent=permanent, blacklist=blacklist, jobtype='realtime', endtime=datetime.now(), dayrange=2, dbinputsensors=sensorlist, basepath=basepath, destination=destination, outputformat=outputformat, recentthreshold=recentthreshold, debug=True)
        statusmsg = filter.apply_filter(db, statusmsg={}, groupdict=groupparameterdict, permanent=permanent, blacklist=blacklist, jobtype='archive', endtime=datetime.now(), dayrange=2, dbinputsensors=sensorlist, basepath=basepath, destination=destination, outputformat=outputformat, recentthreshold=recentthreshold, debug=True)

class TestThreshold(unittest.TestCase):

    def test_assign_parameterlist(self):
        conf = mameth.get_conf(os.path.join('..', 'conf', 'threshold.cfg'))
        para = threshold.assign_parameterlist(threshold.sp.valuenamelist, conf)
        self.assertTrue(para.get('1'))

    def test_get_data(self):
        (dat1,msg1) = threshold.get_data('file', "/tmp/archive/LEMI036_3_0001", "dbcredentials", "LEMI036_3_0001", 1000, startdate=None, debug=True)
        (dat2,msg2) = threshold.get_data('db', "path", "cobsdb", "LEMI036_3_0001", 1000, startdate=None, debug=True)
        self.assertTrue(msg1)

    def test_get_test_value_check_threshold(self):
        data = create_teststream()
        (testvalue, msg2) = threshold.get_test_value(data, key='x', function='average', debug=True)
        self.assertEqual(testvalue, 21000)
        (evaluate, msg) = threshold.check_threshold(testvalue, 20000, "above", debug=True)
        self.assertTrue(evaluate)


    def test_interprete_status(self):
        conf = mameth.get_conf(os.path.join('..', 'conf', 'threshold.cfg'))
        para = threshold.assign_parameterlist(threshold.sp.valuenamelist, conf)
        valuedict = para.get('1', {})
        xxx = threshold.interprete_status(valuedict, debug=True)
        self.assertEqual(xxx, "Current average below 5")


if __name__ == "__main__":
    unittest.main(verbosity=2)


#!/usr/bin/env python
# coding=utf-8

"""
MagPy - Baseline recalculation

This method should be able to do:
- recalculate baselines with or without rotation parameters
- can be run during yearly analysis and update all BLV files on Server
- update DIDATA_XXX raw data tables
- update BLV data in database

        This data is then used in the stepwise procedure:
        (firstrun: uncompensated data run: get average basevalues, get an initial
                    constant baseline estimate, delta F and basis for quality checks)
        (secondrun: compensated data run: apply offsets, apply rotation, get baseline,
                    delta F and basis for quality checks)
        IMPORTANT:
                   1. initial baseline for piers other than A2:
                       -> have delta values (delta D, I, F) been considered?
                       -> which one were considered?
                       Answer: yes! deltaD, deltaI and deltaF are already considered
                               in the absolute method, as long as db is active.
                               Therefor BLV db inputs are already corrected.
                               The input of the PIERS table is read, split by komma,
                               and the last line is used.
                       Problem: each DB line needs a reference to the correction values used.
                               Could be done in DataDeltaValues - NO, because it
                               changes from one day to the other
                               Need to be part of the result line -> string5?
                    2. new delta D and delta I:
                       -> where do I get them?
                       -> how are they calculated?
                       -> at which step are those included?
                       -> how do they compare to the previous values?
                       -> are they written to the database?
                    Resulting application scheme:
                       -> check for delta values in PIERS - those have been used for BLV tables


"""

# Define packges to be used (local refers to test environment)
# ------------------------------------------------------------
from magpy.stream import *
import magpy.absolutes as di
from magpy.core import plot as mp
from magpy.core import database
from magpy.core import methods
from magpy.opt import cred as mpcred

from martas.version import __version__

from dateutil.parser import parse
from shutil import copyfile

import itertools
import getopt
import pwd
import socket
import sys  # for sys.version_info()

from martas.core.methods import martaslog as ml
from martas.core import methods as mm

#from cobs_methods import define_logger, connect_databases, get_primary_instruments, get_stringdate, combinelists, \
#    get_conf


## Methods from yearly_analysis_1 -> can later be imported to yearly_analysis
def baseline_overview(runmode='firstrun', config=None, destpath=None, debug=False):
    """
    DESCRIPTION:
        Provide an overview about existing basevalues. The baseline_overview method will read existing basevalue
        data files from DB or file, named like BLV_vario_scalar_pier. BLV corresponds to blvabb, vario and scalar
        contain the SensorID's of variometer and scalar sensor, pier is obvious.
        After reading the BLV data, flagging info is obtained from DB and the given file path in config. Flagged data
        is dropped. Baseline fits are performed using a spline fit with 0.3 knotstep by default. Plotted are daily means
        of the basevalus.
        In case of "thirdrun":
         - only A2 data is plotted
         - a cleaned BLV_..._YEAR.txt file is written to the given blvdatapath directory
         - the graph is stired to the given figurepath

    APPLICATION:
        Step0: we are reviewing source basevalue data (from DB). All existing flags
        are applied and checked. Further flags are added.

    EXAMPLE:

    """

    print(" ----------------------------------------------------------------------- ")
    print(" ----------------  Cleanup and decide on basevalue data----------------- ")
    print(" --------------------------  for each variometer  ---------------------- ")
    print(" ----------------------------------------------------------------------- ")

    if not config:
        config = 'None'
    # Use all scalars only for firstrun - aftwards use average curve
    vainstlist = config.get('vainstlist')
    scinstlist = config.get('scinstlist')
    pierlist = config.get('pierlist')
    primarypier = config.get('primarypier') #,'A2')
    pierdict = config.get('pierdict') #,'A2')
    year = config.get('year')
    blvabb = config.get('blvabb')
    datasource = config.get('datasource')  # db,file
    blvdatapath = config.get('blvdatapath')
    flagfile = config.get('flagfile')
    db = config.get('primaryDB')
    figurepath = config.get('figurepath', '/tmp')
    knotstep = float(config.get('knotstep', 0.3))

    if not config.get('startdate'):
        starttime = str(year - 1) + '-11-01'
        endtime = str(year + 1) + '-02-01'
    else:
        starttime = config.get('startdate')
        endtime = config.get('enddate')
    symbols = ['d', 's', 'o', 'p', '1', '2', '3', '4', '8', '1', '2']
    ttmp = arange(0, 1, 0.0001)
    legendoffset = 2  # shift data of plot 3 so that legend and data are visible

    if debug:
        print(" - you selected {} as blvabb (baseline data file beginning)".format(blvabb))
        print(" - datasource: {} ".format(datasource))
        print(" - starttime: {} ".format(starttime))
        print(" - endtime: {} ".format(endtime))
        print(" - primaryDB: {} ".format(db))

    if not len(vainstlist) > 0:
        print(" baseline requires variometer data: please provide")
    if not len(scinstlist) > 0:
        print(" baseline requires continuous scalar data: please provide")
    # further required are: sourcepath, outpath
    # sourcepath = sourcepath,'DI','data',
    # sourcepath = outpath,'magpy',
    # flagfile = os.path.join(outpath,'magpy','BLV_flags.json')

    for vainst in vainstlist:
        if runmode == 'thirdrun':
            scinstlist = [scinstlist[0]]
        for scinst in scinstlist:
            for idx, pier in enumerate(pierlist):
                basename = "{}_{}_{}_{}".format(blvabb, vainst[:-5], scinst[:-5], pier)
                if runmode in ['secondrun', 'thirdrun'] and methods.is_number(year):
                    filename = "{}_{}_{}.txt".format(basename, year, runmode)
                else:
                    filename = "{}.txt".format(basename)
                    if debug:
                        print ("Checking for {}".format(basename))
                    if not os.path.isfile(os.path.join(blvdatapath, filename)):
                        filename = "{}_{}_{}.txt".format(basename, year, runmode)

                if debug:
                    print ("Path and Filename", blvdatapath, filename)

                if datasource == 'db':
                    print(" Accessing database...")
                    absr = db.read(basename, starttime=starttime, endtime=endtime)
                else:
                    if os.path.isfile(os.path.join(blvdatapath, filename)):
                        print(" Accessing file...", os.path.join(blvdatapath, filename))
                        if debug:
                            print(" Starttime, Endtime:", starttime, endtime)
                        absr = read(os.path.join(blvdatapath, filename), starttime=starttime, endtime=endtime, datecheck=False)
                        if debug:
                            print(" Obtained {} data points".format(len(absr)))
                    else:
                        absr = DataStream()

                print(" Running analysis for {}".format(basename))

                if absr.length()[0] > 0:
                    print("------------------------------")
                    print("Found {} baseline values for {}".format(absr.length()[0], pier))
                    print("------------------------------")
                    absr = absr.removeduplicates()
                    dataid = absr.header.get('DataID')
                    print(" -- Dropping flags contained in data set:")
                    blvflaglistint = absr.header.get("DataFlags")
                    if blvflaglistint:
                        print ("    found {} flags".format(len(blvflaglistint)))
                        absr = blvflaglistint.apply_flags(absr, mode='drop')
                    if db:
                        print(" -- Dropping flags from DB:", dataid)
                        blvflaglist = db.flags_from_db(dataid)
                        print("   -> {} flags".format(len(blvflaglist)))
                        if len(blvflaglist) > 0:
                            absr = blvflaglist.apply_flags(absr, mode='drop')
                    else:
                        if debug:
                            print ("   No DB connected: Could not access flags from DB")
                    if flagfile:
                        print(" -- Dropping flags from File", flagfile)
                        flaglist = flagging.load(flagfile, sensorid=dataid)
                        print("   -> {} flags".format(len(flaglist)))
                        if len(flaglist) > 0:
                            absr = flaglist.apply_flags(absr, mode='drop')
                    absr = absr._drop_nans('dx')
                    absr = absr._drop_nans('dy')
                    absr = absr._drop_nans('dz')
                    if absr.length()[0] > 10:
                        func = absr.fit(['dx', 'dy', 'dz'], knotstep=knotstep)
                    print(" -- Length after removing nan values: {}".format(absr.length()[0]))
                    blvmeans = absr.dailymeans(keys=['dx', 'dy', 'dz'])
                    blvmeans = blvmeans.sorting()
                    ti = blvmeans._get_column('time')
                    basex = blvmeans._get_column('x')
                    basey = blvmeans._get_column('y')
                    basez = blvmeans._get_column('z')
                    sym = symbols[idx]
                    alpha = 1
                    if pierdict.get(pier):
                        if methods.is_number(pierdict.get(pier)):
                            alpha = float(pierdict.get(pier))
                    plt.subplot(311)
                    plt.title('{}'.format(basename[:-3]))
                    plt.ylabel('BASE H')
                    plt.grid(True)
                    plot(ti, basex, color='darkgray', linestyle='', marker=sym, alpha=alpha)
                    if pier == primarypier:
                        x1, x2, y1, y2 = plt.axis()
                        plt.axis((x1, x2, y1 - 4, y2 + 4))
                    if absr.length()[0] > 10:
                        #print (denormalize(ttmp, func[1], func[2]))
                        if not runmode == 'thirdrun' and len(func) > 0:
                            plot(denormalize(ttmp, func[1], func[2]), func[0]['fdx'](ttmp), 'r-')
                        elif pier == primarypier and len(func) > 0:
                            plot(denormalize(ttmp, func[1], func[2]), func[0]['fdx'](ttmp), 'r-', linewidth=2)
                    plt.subplot(312)
                    # plt.tick_params(axis='y', which='both', labelleft='off', labelright='on')
                    plt.ylabel('BASE D')
                    plt.grid(True)
                    plot(ti, basey, color='darkgray', linestyle='', marker=sym, alpha=alpha)
                    if pier == primarypier:
                        x1, x2, y1, y2 = plt.axis()
                        plt.axis((x1, x2, y1 - 0.05, y2 + 0.05))
                    if absr.length()[0] > 10:
                        if not runmode == 'thirdrun' and len(func) > 0:
                            plot(denormalize(ttmp, func[1], func[2]), func[0]['fdy'](ttmp), 'r-')
                        elif pier == primarypier and len(func) > 0:
                            plot(denormalize(ttmp, func[1], func[2]), func[0]['fdy'](ttmp), 'r-', linewidth=2)
                    plt.subplot(313)
                    plt.ylabel('BASE Z')
                    plt.grid(True)
                    # plt.xticks(rotation='vertical')
                    plot(ti, basez, color='darkgray', linestyle='', marker=sym, alpha=alpha, label=pier)
                    if pier == primarypier:
                        x1, x2, y1, y2 = plt.axis()
                        plt.axis((x1, x2, y1 - 4 - legendoffset, y2 + 4 - legendoffset))
                    if absr.length()[0] > 10:
                        if not runmode == 'thirdrun':
                            plot(denormalize(ttmp, func[1], func[2]), func[0]['fdz'](ttmp), 'r-')
                        elif pier == primarypier:
                            plot(denormalize(ttmp, func[1], func[2]), func[0]['fdz'](ttmp), 'r-', linewidth=2)
                    if absr.length()[0] > 0 and destpath and not debug:
                        print("Writing new abs files to {} as {}".format(destpath, basename))
                        absr.write(destpath, filenamebegins=basename, filenameends='_' + str(year) + '.txt',
                                   format_type='PYSTR', coverage='all')
                else:
                    print("No absolutes found for {} - continuing".format(pier))
            plt.legend(loc="lower left", ncol=3, shadow=True)
            if runmode == 'thirdrun' and vainst == vainstlist[0]:
                plotdir = os.path.join(figurepath, 'Yearbook', 'Graphs')
                print("Saving figure to: (defined by figurepath in cfg)", plotdir)
                if not os.path.exists(plotdir):
                    os.makedirs(plotdir)
                plt.savefig(os.path.join(plotdir, 'allbasevalues.png'))
            plt.show()

    return True


def basevalue_recalc(runmode, config=None, startdate=None, enddate=None, debug=False):
    """
    DESCRIPTION
        recalculate basevalues
    """
    if not config:
        config = {}
    vainstlist = config.get('vainstlist')
    scinstlist = config.get('scinstlist')
    pierlist = config.get('pierlist')
    year = config.get('year')
    plot = config.get('plot', 'False')
    blvabb = config.get('blvabb')
    diindent = config.get('diid', '')
    datasource = config.get('datasource')  # db,file
    blvdatapath = config.get('blvdatapath')
    flagfile = config.get('diflagfile')
    db = config.get('primaryDB')
    obscode = config.get('obscode')
    writemode = config.get('writemode')
    writeflags = config.get('writeflags')
    didatapath = config.get('didatapath')
    backup = config.get('backup', False)
    writeblv2file = config.get('writeblv2file', False)
    writeblv2db = config.get('writeblv2db', False)
    writedi2db = config.get('writedi2db', False)
    fixalpha = config.get('usealpha', True)
    fixbeta = config.get('usebeta', True)
    magrotation = config.get('magrotation', True)
    compensation = config.get('compensation', True)
    expectedD = config.get('expectedD', None)
    expectedI = config.get('expectedI', None)
    skipscalardb = config.get('skipscalardb', False)
    if skipscalardb in ['True','true']:
        skipscalardb = True
    else:
        skipscalardb = False
    movetoarchive = config.get('movetoarchive', '')
    # check_conf determines the archive path
    if not os.path.isdir(movetoarchive):
        if movetoarchive:
            print (" movetoarchive is set but does not seem to be a valid directory")
        movetoarchive = ""
    movetoarchivenow = "" # will be set to path if archiving conditions are found
    contflagfile = config.get('contflagfile', False) # flagging data for vario and scalar from file

    sourcepath = os.path.join(config.get('base'), 'archive', obscode)

    format_type = "PYSTR"

    print("--- RUNNING BASEVALUE CALCULATION ---")

    def _check_db_cond(pier, vario, scalar, conddict):
        if not conddict:
            return True
        # print (conddict)
        boollist = [True]
        vvarios = conddict.get('vario')
        vscalars = conddict.get('scalar')
        vpiers = conddict.get('pier')
        if not isinstance(vvarios, list):
            vvarios = [vvarios]
        if not isinstance(vscalars, list):
            vscalars = [vscalars]
        if not isinstance(vpiers, list):
            vpiers = [vpiers]
        if len(vvarios) > 0 and not vario in vvarios:
            boollist.append(False)
        if len(vpiers) > 0 and not pier in vpiers:
            boollist.append(False)
        if len(vscalars) > 0 and not scalar in vscalars:
            boollist.append(False)

        return np.asarray(boollist).all()

    if not config.get('enddate'):
        startdate = str(year - 1) + '-11-01'
        enddate = str(year + 1) + '-02-01'
    else:
        startdate = config.get('startdate')
        enddate = config.get('enddate')

    if debug:
        print("BLV analysis:")
        print(" blv files will be written to", blvdatapath)

    if runmode in ['thirdrun']:
        if debug:
            print(" Using only primary F as scalar record")
        # it is better to use the primary instrument !!!! -> correct flags are used
        # scinstlist = ['CobsF_sec_0001']
        scinstlist = [scinstlist[0]]

    for scinst in scinstlist:
        if debug:
            print(" Now dealing with scalar {}".format(scinst))
        if scinst:
            if os.path.isdir(scinst):
                print(" scalar instrument provided as full path")
                # directory provided - use this path and get scinst from last layer
                scalarpath = os.path.join(scinst, '*')
                scinst = os.path.basename(os.path.normpath(scinst))
                scalarinst = scinst[:-5]
                print("  -> scalar instrument: {}".format(scinst))
            else:
                scalarinst = scinst[:-5]
                scalarpath = os.path.join(sourcepath, scalarinst, scinst, '*')
            scalarinst = scinst[:-5]
        else:
            scalarpath = None
            scalarinst = "None"
        for vainst in vainstlist:
            if debug:
                print(" Now dealing with vario {}".format(vainst))
            if os.path.isdir(vainst):
                print(" vario instrument provided as full path")
                variopath = os.path.join(vainst, '*')
                vainst = os.path.basename(os.path.normpath(vainst))
                print("  -> vario instrument: {}".format(vainst))
            else:
                variopath = os.path.join(sourcepath, vainst[:-5], vainst, '*')
            if fixalpha and db:
                # IMPORTANT: use alpha for current year
                header = db.fields_to_dict(vainst)
                exist = header.get('DataRotationAlpha', '').split(',')
                rotdic = {}
                for el in exist:
                    cont = el.split('_')
                    try:
                        rotdic[int(cont[0])] = float(cont[1])
                    except:
                        pass
                alpha = rotdic.get(year, 0.0)
                if alpha == 0:
                    print(" - ALPHA is zero - check if last year is existing")
                    alpha = rotdic.get(year - 1, 0.0)
                    if alpha == 0:
                        print(" - ALPHA not existing yet - setting alpha=0")
                    else:
                        print(" - using value for {}: {}".format(year - 1, alpha))
            else:
                alpha = None

            if fixbeta and db:
                # IMPORTANT: use alpha for current year
                header = db.fields_to_dict(vainst)
                exist = header.get('DataRotationBeta', '').split(',')
                rotdic = {}
                for el in exist:
                    cont = el.split('_')
                    try:
                        rotdic[int(cont[0])] = float(cont[1])
                    except:
                        pass
                beta = rotdic.get(year, 0.0)
                if beta == 0:
                    print(" - BETA is zero - check if last year is existing")
                    beta = rotdic.get(year - 1, 0.0)
                    if beta == 0:
                        print("BETA not existing yet - setting beta=0")
                    else:
                        print(" - using value for {}: {}".format(year - 1, beta))
            else:
                beta = None

            if debug:
                print(" rotation data determined")

            if debug:
                # contains a path
                print("MOVETOARCHIVE looks like ", movetoarchive)
            if movetoarchive:
                # only activate for last vario and last secular
                lastsc = scinstlist[-1]
                lastva = vainstlist[-1]
                # print (" LAST instruments:", lastsc, lastva)
                # print (" and the insts", scinst, vainst)
                if lastsc.find(scinst) >= 0 and lastva.find(vainst) >= 0:
                    print(" -> activating movetoarchive")
                    movetoarchivenow = movetoarchive
                else:
                    movetoarchivenow = ""

            for pier in pierlist:
                if debug:
                    print(" Analyzing for pier {}".format(pier))
                if diindent in ['All','all','ALL']:
                    blvid = ".txt"
                elif diindent:
                    print ("Got DIID:", diindent)
                    blvid = diindent
                else:
                    blvid = "{}_{}.txt".format(pier, obscode)
                skipscalardb = False
                if runmode in ['secondrun', 'firstrun']:
                    deltaD = 0.0000000001
                    deltaI = 0.0000000001
                else:
                    # skipscalardb = True
                    # scalarinst = 'CobsF'
                    # scalarpath = os.path.join(outpath,'magpy','CobsF_sec*')
                    deltaD = None
                    deltaI = None
                if pier == 'A16':
                    abstype = 'autodif'
                    azimuth = 267.36651
                else:
                    abstype = 'manual'
                    azimuth = False
                if debug:
                    print(" di data provided in {}".format(didatapath))
                print("----------------------------------------------------------")
                print("----------------------------------------------------------")
                print("{} absolute analysis for pier {} with {} and {}".format(year, pier, vainst[:-5], scalarinst))
                print("----------------------------------------------------------")
                print("----------------------------------------------------------")
                # absstream = absoluteAnalysis(stationid=stationid, deltaF=deltaF)
                dbadd = False
                if writedi2db and not debug:
                    # writedi2db only supports two write modes: append/insert and fullreplace
                    dbadd = True
                    if writemode == 'fullreplace':
                        prepare_table(db, "DIDATA_{}".format(obscode), startdate, enddate, tcol='StartTime',
                                      cond={'DIID': pier})
                    elif writemode in ['replace', 'overwrite']:
                        print(" writedi2db only supports two write modes: append/insert and fullreplace")

                if debug:
                    print("Parameter", didatapath, variopath, scalarpath, startdate, enddate)
                    print("Parameter", db, skipscalardb, compensation, dbadd, azimuth, magrotation)
                    print("Parameter", alpha, beta, deltaD, deltaI)
                absresult = di.absolute_analysis(didatapath, variopath, scalarpath, diid=blvid, pier=pier,
                                                 expD=expectedD, expI=expectedI, starttime=startdate, endtime=enddate,
                                                 db=db, skipscalardb=skipscalardb, compensation=compensation,
                                                 magrotation=magrotation, alpha=alpha, beta=beta, abstype=abstype,
                                                 azimuth=azimuth, deltaD=deltaD, deltaI=deltaI, dbadd=dbadd,
                                                 movetoarchive=movetoarchivenow, flagfile=contflagfile, debug=debug)
                # TODO addDB
                # How is addDB working? -> Append new DI files? Replace existing?
                # How to clean files within the time range?

                if absresult and absresult.length()[0] > 0:
                    absresult = absresult.removeduplicates()
                    # remove all flagged data set (with CobsF)
                    basename = '{}_{}_{}_{}'.format(blvabb, vainst[:-5], scalarinst, pier)  # too save
                    flagname = '{}_{}_{}_{}'.format('BLV', vainst[:-5], scalarinst, pier)  # as flags are stored
                    absresult.header["SensorID"] = flagname
                    absresult.header["DataID"] = flagname
                    print(" -------------------------------------")
                    print(" Storing data with SensorID {}".format(flagname))
                    print(" -------------------------------------")
                    print(" - getting flags:")
                    flaglist = []
                    if db:
                        flaglist = db.flags_from_db(flagname)
                        print("   -> {} flags in db".format(len(flaglist)))
                    fllist = flagging.load(flagfile, sensorid=flagname)
                    print("   -> {} flags in file {}".format(len(fllist), flagfile))
                    if len(fllist) > 0 and len(flaglist) > 0:
                        flaglist = flaglist.join(fllist)
                        # flaglist.extend(fllist)
                    elif not len(flaglist) > 0:
                        flaglist = fllist
                    if len(flaglist) > 0:
                        # flaglist = absresult.flaglistclean(flaglist)
                        print("  -> Applying {} flags".format(len(flaglist)))
                        if writeflags:
                            absresult = flaglist.apply_flags(absresult, mode='insert')
                            # TODO or absresult.header['DataFlags'] = flaglist
                            # absresult = absresult.flag(flaglist)
                        else:
                            print("  -> Dropping flagged data")
                            absresult = flaglist.apply_flags(absresult, mode='drop')
                            # absresult = absresult.remove_flagged()
                    if not runmode in ['firstrun', 'secondrun', 'thirdrun']:
                        filenamebegins = "{}".format(basename)
                    else:
                        filenamebegins = "{}_{}_{}".format(basename, year, runmode)

                    extension = '.txt'
                    if format_type == 'PYCDF':
                        extension = '.cdf'

                    if writeblv2file:
                        succ = True
                        if backup:
                            succ = backup_file(os.path.join(blvdatapath, filenamebegins + extension))
                        if succ:
                            print(" - writing data to file using writemode {} ...".format(writemode))
                            wm = writemode
                            if wm in ['append', 'skip']:
                                wm = 'skip'
                            if wm == 'fullreplace':
                                wm = prepare_file(os.path.join(blvdatapath, filenamebegins + extension), startdate,
                                                  enddate, format_type=format_type)
                            print("   -> writing {} data points".format(absresult.length()[0]))
                            absresult.write(blvdatapath, coverage='all', filenamebegins=filenamebegins,
                                            format_type=format_type, mode=wm)
                            print(" ... success")

                    if writeblv2db and _check_db_cond(pier, vainst, scinst, config.get('writedbcond')):
                        print(" - Writing data to DB using writemode {} ...".format(writemode))
                        wm = writemode
                        if wm in ['overwrite', 'delete']:
                            wm = 'delete'
                        if wm in ['append', 'skip']:
                            wm = 'insert'
                        if wm == 'fullreplace':
                            wm = prepare_table(db, filenamebegins, startdate, enddate)
                        # - Replace everything within time range (fullreplace) (not existing in file)
                        # - Replace only existing data (replace) - (replace in file)
                        # - Append data if not existing (append) - (skip in file)
                        # - Delete all and write new file (overwrite) - (overwrite in file)
                        if db:
                            db.write(absresult, tablename=filenamebegins, mode=wm)
                        print(" ... success")

                    if plot == 'True':
                        try:
                            absresult = absresult._drop_nans('dx')
                            absresult = absresult._drop_nans('dy')
                            absresult = absresult._drop_nans('dz')
                            func = absresult.fit(['dx', 'dy', 'dz'], fitfunc='spline', knotstep=0.3)
                            mp.tsplot([absresult], [['dx', 'dy', 'dz']], symbols=[['o', 'o', 'o']],
                                      padding=[5, 0.005, 5], functions=[[func,func,func]], title=pier)
                        except:
                            print(" - Not enough data points for suitable baseline")
                            pass
                    print(" saving (and eventually plotting) successfully finished ")
                    print(" -------------------------------------")
                else:
                    print(" -> obtained an empty abs-result")
                    print(" -------------------------------------")


def check_conf(config, startdate, enddate, varios=None, scalars=None, piers=None, debug=False):
    """
    DESCRIPTION
        this method will extend/modify the configuration data with basevalue analysis specific parameters
    """

    if varios == None:
        varios = []
    if scalars == None:
        scalars = []
    if piers == None:
        piers = []
    if enddate == 'now' or config.get('enddate') == 'now':
        endtime = datetime.now(timezone.utc).replace(tzinfo=None)
        config['enddate'] = endtime.strftime("%Y-%m-%d")
        if not startdate:
            starttime = endtime - timedelta(days=380)
            config['startdate'] = starttime.strftime("%Y-%m-%d")
    elif enddate == 'file':
        # determine startdate and enddate from the available files in the directory
        pass
    elif enddate:
        endtime = parse(enddate)
        config['enddate'] = endtime.strftime("%Y-%m-%d")
        if not startdate:
            starttime = endtime - timedelta(days=380)
            config['startdate'] = starttime.strftime("%Y-%m-%d")
    else:
        enddate = config.get('enddate', None)
        if enddate in ['None', 'False', None, 0]:
            config['enddate'] = None
    if startdate:
        starttime = parse(startdate)
        config['startdate'] = starttime.strftime("%Y-%m-%d")
    else:
        startdate = config.get('startdate', None)
        if startdate in ['None', 'False', None, 0]:
            config['startdate'] = None

    if varios and len(varios) > 0:
        print(" Setting variometers to {}".format(varios))
        config['vainstlist'] = varios
    if scalars and len(scalars) > 0:
        print(" Setting scalars to {}".format(scalars))
        config['scinstlist'] = scalars
    if piers and len(piers) > 0:
        print(" Setting piers to {}".format(piers))
        config['pierlist'] = piers
    pl = config.get('pierlist')
    if not isinstance(pl, (list, tuple)):
        config['pierlist'] = [pl]
    if not config.get('primarypier'):
        if len(config.get('pierlist')) > 0:
            config['primarypier'] = config.get('pierlist')[0]

    if config.get('year') in ['None', 'False', None, 0]:
        config['year'] = None
    elif config.get('year') in ['current', 'now']:
        config['year'] = datetime.now(timezone.utc).replace(tzinfo=None).year

    if config.get('blvdatapath', '') in ['None', 'False', '', None]:
        config['blvdatapath'] = os.path.join(config.get('base', ''), 'archive', config.get('obscode', ''), 'DI', 'data')

    if not isinstance(config.get('dbcredentials'), list):
        config['dbcredentials'] = [config.get('dbcredentials')]
    if not isinstance(config.get('vainstlist'), list):
        config['vainstlist'] = [config.get('vainstlist')]
    if not isinstance(config.get('scinstlist'), list):
        config['scinstlist'] = [config.get('scinstlist')]
    if not isinstance(config.get('pierlist'), list):
        config['pierlist'] = [config.get('pierlist')]

    credentials = config.get('dbcredentials')
    credential = None
    if len(credentials) > 0:
        credential = credentials[0]
    if debug:
        print("- Connecting to database {}".format(credential))
    try:
        db = mm.connect_db(credential)
        print (" ... success")
        config['primaryDB'] = db
    except:
        config['primaryDB'] = None

    sourcepath = os.path.join(config.get('base', ''), 'archive', config.get('obscode', ''))

    if config.get('didatapath') == 'raw':
        config['didatapath'] = os.path.join(sourcepath, 'DI', 'raw')
    elif config.get('didatapath') == 'analyze':
        config['didatapath'] = os.path.join(sourcepath, 'DI', 'analyze')
        if config.get('movetoarchive') in ['True', 'true', True]:
            config['movetoarchive'] = os.path.join(sourcepath, 'DI', 'raw')
    elif config.get('didatapath') == '':
        config['didatapath'] = os.path.join(sourcepath, 'DI', 'raw')

    for key in config:
        if config.get(key) in ['False', 'false']:
            config[key] = False
        if config.get(key) in ['True', 'true']:
            config[key] = True

    return config


def get_runmode(joblist):
    if 'firstrun' in joblist:
        runmode = 'firstrun'
        if not 'overview' in joblist:
            joblist.append('default')
    elif 'secondrun' in joblist:
        runmode = 'secondrun'
        if not 'overview' in joblist:
            joblist.append('default')
    elif 'thirdrun' in joblist:
        runmode = 'thirdrun'
        if not 'overview' in joblist:
            joblist.append('default')
    else:
        runmode = 'upload'
        if not 'overview' in joblist:
            joblist.append('default')
    return list(set(joblist)), runmode


def backup_file(path):
    ret = False
    now = datetime.now(timezone.utc).replace(tzinfo=None)
    agecode = now.strftime("%Y%m%d%H")
    print(" Backing up {}".format(path))
    if os.path.isfile(path):
        directory, finame = os.path.split(path)
        newfilename = "{}.{}.bak".format(finame, agecode)
        newpath = os.path.join(directory, newfilename)
        try:
            shutil.copyfile(path, newpath)
            print(" Backup file saved as {}".format(newpath))
            ret = True
        except:
            pass
    else:
        print("File not yet existing")
        ret = True
    return ret


def prepare_file(path, starttime, endtime, format_type='PYSTR'):
    """
    DESCRIPTION
        Prepare file will remove all contents from the file within the selected timerange
    """
    print(" -> fullreplace: preparing file")
    try:
        # read data
        data = read(path)
        # remove time range
        data1 = data.trim(endtime=starttime)
        data2 = data.trim(starttime=endtime)
        newdata = join_streams(data1, data2)
        # write data
        directory, finame = os.path.split(path)
        fname = finame.split('.')
        newdata.write(directory, filenamebegins=fname[0], coverage='all', format_type=format_type, mode='overwrite')
    except:
        print("File not yet existing?")
    # set wm to replace
    return 'replace'


def prepare_table(db, tablename, starttime, endtime, tcol='time', cond=None):
    # starttime and endtime need to be strings
    if not cond:
        cond = {}
    print(" -> fullreplace: preparing table")
    condition = ''
    if cond:
        for key in cond:
            condstr = "{} LIKE '{}%' AND ".format(key, cond[key])
            condition += condstr
    try:
        delstring = "DELETE FROM {} WHERE {}DATE({}) BETWEEN DATE('{}') AND DATE('{}')".format(tablename, condition,
                                                                                               tcol, starttime, endtime)
        cursor = db.db.cursor()
        cursor.execute(delstring)
        cursor.close()
    except:
        print(" PrepareTable: removing time range failed - table not yet existing?")
    return 'replace'


def main(argv):
    version = __version__
    configpath = ''
    statusmsg = {}
    debug = False
    joblist = ['default']
    startdate = None
    enddate = None
    scalars = []
    varios = []
    piers = []
    proxies = {}

    # supported jobs are:
    #   (1) calculate basevalues from DI data and write BLV files/DB
    #           -> caluclate BLV files from DI data - daily basis
    #                 (two confs: one for comp with archive, one without comp)
    #           -> redo any BLV analsysis for a specified time range with
    #                 given instruments
    #   (2) homogenize DB contents and files
    #   (3) create overview plots

    try:
        opts, args = getopt.getopt(argv, "hc:j:s:e:v:f:p:D",
                                   ["config=", "joblist=", "startdate=", "enddate=", "variometers=", "scalars=",
                                    "piers=", "debug="])
    except getopt.GetoptError:
        print('basevalue.py -c <config>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('-------------------------------------')
            print('Description:')
            print('-- basevalue.py will obtain baseline plots --')
            print('-----------------------------------------------------------------')
            print('detailed description:')
            print(' basevalue.py recalculates basevalues from DI measurements and provided variation and scalar data.')
            print(' The method can use multiple data sources and piers as defined in the configuration file. It')
            print(' further supports different run modes defining the complexity of baseline fits, application of ')
            print(' rotation matricies etc. These run modes are used for the yearly definitive data analysis of the')
            print(' Conrad Obs. It is recommended to use a similar data coverage of approximately one year when using')
            print(' this technique with polynomial or spline fits to get comparable fitting parameters.')
            print(' In default mode: if enddate is set to now and no startdate is given, then startdate')
            print(' is set to now-380days')
            print('')
            print('-------------------------------------')
            print('Usage:')
            print('python basevalue.py -c <config>')
            print('-------------------------------------')
            print('Options:')
            print('-c (required) : configuration data path')
            print("-j            : list of jobs - one of 'firstrun','seconrun','thirdrun','upload'")
            print("              : plus 'default' or 'overview'")
            print("              : -j firstrun,overview will create overview plots using firstrun fits")
            print('-e            : enddate (e.g. now, 2020-11-22)')
            print('-s            : startdate')
            print('-v            : variometer')
            print('-f            : scalar')
            print('-p            : piers')
            print('-------------------------------------')
            print('Application:')
            print(
                'python3 basevalue.py -c /media/leon/Images/products/data/magnetism/definitive/wic2020/basevalue.cfg -j upload -s 2019-01-01 -e 2021-02-01')
            sys.exit()
        elif opt in ("-c", "--config"):
            # delete any / at the end of the string
            configpath = os.path.abspath(arg)
        elif opt in ("-e", "--enddate"):
            enddate = arg
        elif opt in ("-s", "--startdate"):
            startdate = arg
        elif opt in ("-j", "--joblist"):
            joblist = arg.split(',')
        elif opt in ("-v", "--variometers"):
            varios = arg.split(',')
        elif opt in ("-f", "--scalars"):
            scalars = arg.split(',')
        elif opt in ("-p", "--piers"):
            piers = arg.split(',')
        elif opt in ("-D", "--debug"):
            # delete any / at the end of the string
            debug = True

    print("Running basevalue version {}".format(version))
    print("--------------------------------")

    if not os.path.exists(configpath):
        print('!! Specify a valid path to configuration information !!')
        print('-- check basevalue.py -h for more options and requirements')
        sys.exit()

    if debug:
        print("- Read and check validity of configuration data")
    config = mm.get_conf(configpath)
    config = check_conf(config, startdate, enddate, varios=varios, scalars=scalars, piers=piers, debug=debug)
    if debug:
        print("  Configuration data:", config)

    #print("2. Activate logging scheme as selected in config")
    #config = define_logger(config=config, category="Info", job=os.path.basename(__file__),
    #                       newname='mm-basevalue-tool.log', debug=debug)
    # Use the logfile from config
    name1 = "{}".format(os.path.basename(config.get('logfile')))
    name1 = name1.replace(".log","")
    statusmsg[name1] = 'Baseline notification successful'
    if not config.get('primaryDB'):
        statusmsg[name1] = 'database failed'

    if debug:
        print("- Running data analysis")
    joblist, runmode = get_runmode(joblist)
    if debug:
        print("  -> you selected the following jobs ({}) and runmode {}".format(joblist, runmode))

    if 'default' in joblist:
        print("Running default job")
        if debug:
            print("Selected parameters:")
            print(runmode, config.get('usealpha'), config.get('magrotation'), config.get('compensation'), startdate,
                  enddate, config.get('year'))
        basevalue_recalc(runmode, config=config, debug=debug)

    if 'overview' in joblist:
        print("Creating plots and overview")
        if debug:
            print("Selected parameters:")
            print(runmode, startdate, enddate, config.get('year'))
        baseline_overview(runmode=runmode, config=config, debug=debug)

    print("basevalue successfully finished")

    # 6. Logging section
    # ###########################
    if not debug:
        receiver = config.get('notification')
        notificationcfg = config.get('notificationconfig')
        if config.get('https'):
            proxies['https'] = config.get('https')
        if config.get('http'):
            proxies['http'] = config.get('http')
        martaslog = ml(logfile=config.get('logfile'), receiver=receiver, proxies=proxies)
        if receiver == 'telegram':
            martaslog.telegram['config'] = notificationcfg
        elif receiver == 'email':
            martaslog.email['config'] = notificationcfg
        martaslog.msg(statusmsg)
    else:
        print("Debug selected - statusmsg looks like:")
        print(statusmsg)


if __name__ == "__main__":
    main(sys.argv[1:])

"""
### Configuration file for DI and Basevalue analysis

### Some general hints for parameter settings:
### (1) update yearly analysis data in server tables and DB
###   -> update raw file directory with current files
###   -> writemode=fullreplace,writeblv=True,writedi=True
### (2) local yearly analysis
###   -> writemode=fullreplace,writeblv2db=False,writedi2db=False
### (3) permanenty running with rotation
###   -> writemode=replace,didatapath=analyze,backup=False
### (4) permanenty running without rotation
###   -> writemode=replace,didatapath=analyze,backup=False
###   -> fixalpha=False,compensation=True,magrotation=False
###   -> pierlist=A2

### Basically run several jobs: 
### 1) A2, all instruments save DB and file
### 2) other piers, all instr, save file (eventually only primaryinst)
### 3) A2, all instr, no rotation, save file (eventually only primaryinst)


# Timerange
# --------------------------------------------------------------
#  - if a year is provided (e.g. 2020), then start and time will 
     be set to 1.11.2019 prev year and 1.2.2021 
#    -> can be overruled by startdate and enddate as options
# --------------------------------------------------------------
year           :      current
#startdate      :      None
#enddate        :      None


# Instruments
# --------------------------------------------------------------
#  - IMPORTANT: Lists need to be provided. If only one instrument
#             or pier then end with comma
#  - reference is a single instrument - provide SensorID
# --------------------------------------------------------------
#vainstlist     :      LEMI036_1_0002_0002,LEMI025_22_0003_0002
#scinstlist     :      GP20S3NSS2_012201_0001_0001,GSM90_14245_0002_0002,GP20S3NSS3_012201_0001_0001
vainstlist     :      LEMI036_1_0002_0002
scinstlist     :      GP20S3NSS2_012201_0001_0001
#pierlist       :      A2,A7,H1,A16,A4,A8,A10,A5,A1
pierlist       :      A5
mobileinst     :      GSM90_31968_0002


# Databases
# --------------------------------------------------------------
dbcredentials  :      cobsdb


# Analysis parameters
# --------------------------------------------------------------
# - magrotation = True will activate compensation and rotation 
# - compensation = True will activate compensation and rotation 
# - usealpha will handle rotation parameter
# - skipscalardb True will not use offsets from database
#    -> use only if your scalar source is fully corrected for 
#       pier diffs already TODO OR TOGETHER WITH F-ADOPTION
# - movetoarchive: if True and didatapath = analyze then
#      successfully analyzed DI data is moved to raw
# --------------------------------------------------------------
magrotation    :      True
compensation   :      True
usealpha       :      True
expectedD      :      4
expectedI      :      64
skipscalardb   :      False
movetoarchive  :      False


# Paths and data sources
# --------------------------------------------------------------
#  - basepath expects a "magpy" directory structure below
#    like /srv    -> /srv/archive/WIC/Di; /srv/archive/WIC/
#  - if blvdatapath = None: /srv/archive/obscode/DI/data 
#  - didata path can be either analyze, raw or a full path
#  - backup = True will create a backup of BLV data before saving
#  - writemode supports:
#          -> replace: Replace only existing data
#          -> overwrite: Delete all and write new
#          -> append: Append data if not existing
#          -> fullreplace: Replace everything within time range
# --------------------------------------------------------------
obscode        :      WIC
datasource     :      file
ext            :      cdf
base           :      /media/leon/Images
plot           :      False
blvabb         :      BLVcomp
#blvdatapath    :      /media/leon/Images/products/data/magnetism/definitive/wic2020/magpy
flagfile       :      /media/leon/Images/products/data/magnetism/definitive/wic2020/magpy/BLV_flags.json
#saveplotpath   :      xxx
didatapath     :      raw
writemode      :      fullreplace
backup         :      True


# Data receivers
# --------------------------------------------------------------
#  - writeflags: write flags within the DataStream structure
# --------------------------------------------------------------
writeraw       :      False
writeblv2db    :      True
writeblv2file  :      False
writedi2db     :      True
writeflags     :      False
writedbcond    :      vario:LEMI036_1_0002_0002;scalar:GP20S3NSS2_012201_0001_0001;pier:A5

# Monitoring
# --------------------------------------------------------------
#  - Logfile (a json style dictionary, which contains statusmessages) 
#  - Notifaction (uses martaslog class, one of email, telegram, mqtt, log) 
# --------------------------------------------------------------
loggingdirectory       :   /var/log/magpy
notification         :   telegram
notificationconfig   :   /myconfpath/mynotificationtype.cfg

"""


